/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <string.h> /* strerror() */
#include <strings.h> /* strcasecmp() */

#include "3rdparty/utlist.h" /* DL_*() */
#include "dnsbl/domain.h" /* decls for own functions */
#include "dnsbl/match.h" /* struct acopm_dnsbl_match, acopm_dnsbl_match_free() */
#include "dnsbl/internal.h" /* internal decls */
#include "utils/log.h" /* acopm_log_*() */
#include "utils/siphash24.h" /* HASH_ITER() */

struct acopm_dnsbl_domain *
acopm_dnsbl_domain_new(struct acopm_dnsbl_client *const restrict client)
{
	struct acopm_dnsbl_domain *const domain = calloc(1, sizeof *domain);
	if (! domain)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return NULL;
	}

	domain->client = client;

	(void) acopm_log_debug("%s: created domain", __func__);
	return domain;
}

void
acopm_dnsbl_domain_set_type(struct acopm_dnsbl_domain *const restrict domain, const enum acopm_dnsbl_aftype type)
{
	assert(domain != NULL);

	domain->type = type;
}

void
acopm_dnsbl_domain_set_host(struct acopm_dnsbl_domain *const restrict domain, const char *const restrict host)
{
	assert(domain != NULL);
	assert(host != NULL);

	if (! acopm_strset_buf(domain->host, host))
		(void) acopm_log_warning("%s: host '%s' too long, truncated", __func__, host);
}

bool
acopm_dnsbl_domain_add(struct acopm_dnsbl_domain *const restrict domain)
{
	assert(domain != NULL);
	assert(domain->client != NULL);

	struct acopm_dnsbl_client *const client = domain->client;

	if (! *domain->host)
	{
		(void) acopm_log_error("%s: attempt to add non-existent domain to client", __func__);
		(void) acopm_dnsbl_domain_free(domain);
		return false;
	}
	if (client->domains)
	{
		struct acopm_dnsbl_domain *tmp;
		DL_FOREACH(client->domains, tmp)
		{
			if (strcasecmp(tmp->host, domain->host) == 0)
			{
				(void) acopm_log_error("%s: attempt to add duplicate domain %s to client",
				                       __func__, domain->host);

				(void) acopm_dnsbl_domain_free(domain);
				return false;
			}
		}
	}

	(void) acopm_log_debug("%s: adding domain to client", __func__);
	DL_APPEND(client->domains, domain);
	return true;
}

void
acopm_dnsbl_domain_free(struct acopm_dnsbl_domain *const restrict domain)
{
	assert(domain != NULL);
	assert(domain->client != NULL);

	struct acopm_dnsbl_client *const client = domain->client;

	(void) acopm_log_debug("%s: freeing domain from client", __func__);

	if (domain->matches)
	{
		struct acopm_dnsbl_match *match, *tmp;
		HASH_ITER(hh, domain->matches, match, tmp)
		{
			if (! domain->matches)
				break;

			(void) acopm_dnsbl_match_free(match);
		}
	}

	if (client->domains)
		DL_DELETE(client->domains, domain);

	(void) free(domain);
}
