/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <string.h> /* strerror() */

#include <arpa/inet.h> /* inet_pton() */
#include <sys/socket.h> /* AF_INET, AF_INET6 */

#include "3rdparty/utlist.h" /* DL_*() */
#include "dnsbl/lookup.h" /* decls for own functions */
#include "dnsbl/request.h" /* acopm_dnsbl_request_*() */
#include "dnsbl/internal.h" /* internal decls */
#include "utils/log.h" /* acopm_log_*() */

// NOTE TO SELF: DO NOT INLINE THIS FUNCTION (Clang can do it; but GCC cannot, due to vararg list)
static bool acopm_attr_printf(3, 4)
acopm_dnsbl_add_str(char *const restrict str, size_t *const restrict offset, const char *const restrict format, ...)
{
	const size_t off = ((offset) ? (*offset) : (0));

	va_list argv;
	va_start(argv, format);
	const int ret = vsprintf(str + off, format, argv);
	va_end(argv);

	if (offset && ret > 0)
		*offset += (size_t) ret;
	else if (offset)
		*offset = 0;

	return ((ret >= 0) ? (true) : (false));
}

struct acopm_dnsbl_lookup *
acopm_dnsbl_lookup_new(struct acopm_dnsbl_client *const restrict client, const char *const restrict address,
                       const acopm_dnsbl_callback callback)
{
	assert(client != NULL);
	assert(client->domains != NULL);
	assert(address != NULL);
	assert(callback != NULL);

	if (! client->ready)
		return NULL;

	struct acopm_dnsbl_lookup *const lookup = calloc(1, sizeof *lookup);
	if (! lookup)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return NULL;
	}

	if (! acopm_strset_buf(lookup->address, address))
	{
		(void) acopm_log_error("%s: address '%s' too long", __func__, address);
		(void) free(lookup);
		return NULL;
	}

	lookup->client = client;
	lookup->callback = callback;

	unsigned char address_oct[16];
	if (
		inet_pton((lookup->family = AF_INET),  address, address_oct) != 1 &&
		inet_pton((lookup->family = AF_INET6), address, address_oct) != 1
	) {
		(void) acopm_log_error("%s: not a valid IP address: %s", __func__, address);
		(void) free(lookup);
		return NULL;
	}

	struct acopm_dnsbl_domain *domain = NULL;

	bool match = false;

	DL_FOREACH(client->domains, domain)
	{
		if (lookup->family == AF_INET && ! (domain->type & ACOPM_DNSBL_AFTYPE_IPV4))
			continue;

		if (lookup->family == AF_INET6 && ! (domain->type & ACOPM_DNSBL_AFTYPE_IPV6))
			continue;

		match = true;
		break;
	}

	if (! match)
	{
		(void) acopm_log_warning("%s: no DNSBLs added for address family", __func__);
		(void) free(lookup);
		return NULL;
	}

	size_t offset = 0;

	if (lookup->family == AF_INET)
	{
		for (size_t x = INET_ADDRLEN; x; x--)
			if (! acopm_dnsbl_add_str(lookup->address_rev, &offset, "%u.", address_oct[x - 1]))
				break;
	}
	else
	{
		for (size_t x = INET6_ADDRLEN; x; x--)
			if (! acopm_dnsbl_add_str(lookup->address_rev, &offset, "%x.%x.",
			                          (unsigned int) (address_oct[x - 1] & 0x0F),
			                          (unsigned int) ((address_oct[x - 1] & 0xF0) >> 4)))
				break;
	}

	if (! offset)
	{
		(void) acopm_log_error("%s: vsprintf: %s", __func__, strerror(errno));
		(void) free(lookup);
		return NULL;
	}

	bool success = false;
	DL_FOREACH(client->domains, domain)
		if (acopm_dnsbl_request_new(lookup, domain))
			success = true;

	if (! success)
	{
		(void) acopm_dnsbl_lookup_free(lookup);
		return NULL;
	}

	(void) acopm_log_debug("%s: looking up %s", __func__, address);
	DL_APPEND(client->lookups, lookup);
	return lookup;
}

void
acopm_dnsbl_lookup_free(struct acopm_dnsbl_lookup *const restrict lookup)
{
	assert(lookup != NULL);

	if (lookup->requests)
	{
		struct acopm_dnsbl_request *request, *tmp;
		DL_FOREACH_SAFE(lookup->requests, request, tmp)
		{
			if (! lookup->requests)
				break;

			(void) acopm_dnsbl_request_free(request);
		}
	}
	else
	{
		(void) acopm_log_debug("%s: freeing lookup on client", __func__);

		struct acopm_dnsbl_client *const client = lookup->client;

		if (client->lookups)
			DL_DELETE(client->lookups, lookup);

		(void) free(lookup);
	}
}
