/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <string.h> /* strerror() */

#include <arpa/inet.h> /* inet_ntop() */
#include <sys/socket.h> /* AF_INET, AF_INET6 */

#include "dnsbl/match.h" /* decls for own functions */
#include "dnsbl/internal.h" /* internal decls */
#include "utils/log.h" /* acopm_log_*() */
#include "utils/siphash24.h" /* HASH_*() */

struct acopm_dnsbl_match *
acopm_dnsbl_match_new(struct acopm_dnsbl_domain *const restrict domain)
{
	struct acopm_dnsbl_match *const match = calloc(1, sizeof *match);
	if (! match)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return NULL;
	}

	match->domain = domain;

	(void) acopm_log_debug("%s: created match", __func__);
	return match;
}

void
acopm_dnsbl_match_set_addr(struct acopm_dnsbl_match *const restrict match, const uint32_t addr)
{
	assert(match != NULL);

	char addr_str[INET_ADDRSTRLEN];
	if (! inet_ntop(AF_INET, &addr, addr_str, sizeof addr_str))
		(void) acopm_log_error("%s: inet_ntop: %s", __func__, strerror(errno));
	else
		(void) acopm_log_debug("%s: setting address to %s", __func__, addr_str);

	match->addr = addr;
}

void
acopm_dnsbl_match_set_desc(struct acopm_dnsbl_match *const restrict match, const char *const restrict desc)
{
	assert(match != NULL);

	(void) acopm_log_debug("%s: setting description to %s", __func__, desc);

	if (! acopm_strset_buf(match->desc, desc))
		(void) acopm_log_warning("%s: description '%s' too long, truncated", __func__, desc);
}

bool
acopm_dnsbl_match_add(struct acopm_dnsbl_match *const restrict match)
{
	assert(match != NULL);
	assert(match->domain != NULL);

	struct acopm_dnsbl_domain *const domain = match->domain;

	if (! match->addr)
	{
		(void) acopm_log_error("%s: attempt to add non-existent match to domain", __func__);
		return false;
	}

	char addr_str[INET_ADDRSTRLEN];
	if (! inet_ntop(AF_INET, &match->addr, addr_str, sizeof addr_str))
	{
		(void) acopm_log_error("%s: inet_ntop: %s", __func__, strerror(errno));
		return false;
	}

	if (domain->matches)
	{
		struct acopm_dnsbl_match *tmp;
		HASH_FIND(hh, domain->matches, &match->addr, sizeof match->addr, tmp);
		if (tmp)
		{
			assert(tmp->addr == match->addr);
			(void) acopm_log_error("%s: attempt to add duplicate match %s to domain", __func__, addr_str);
			return false;
		}
	}

	(void) acopm_log_debug("%s: adding match to domain", __func__);
	HASH_ADD(hh, domain->matches, addr, sizeof match->addr, match);
	return true;
}

void
acopm_dnsbl_match_free(struct acopm_dnsbl_match *const restrict match)
{
	assert(match != NULL);
	assert(match->domain != NULL);

	(void) acopm_log_debug("%s: freeing match", __func__);

	struct acopm_dnsbl_domain *const domain = match->domain;

	if (domain->matches)
		HASH_DELETE(hh, domain->matches, match);

	(void) free(match);
}
