/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <string.h> /* strerror() */

#include <arpa/inet.h> /* inet_ntop() */
#include <sys/socket.h> /* AF_INET */

#include <event2/dns.h> /* evdns_*() */

#include "3rdparty/utlist.h" /* DL_*() */
#include "dnsbl/lookup.h" /* acopm_dnsbl_lookup_free() */
#include "dnsbl/request.h" /* decls for own functions */
#include "dnsbl/internal.h" /* internal decls */
#include "utils/log.h" /* acopm_log_*() */
#include "utils/siphash24.h" /* HASH_FIND() */

static void
acopm_dnsbl_resolve_callback(const int code, const char type, const int count, const int ttl,
                             void *const restrict result, void *const restrict data)
{
	(void) ttl;

	if (code == DNS_ERR_CANCEL)
		return;

	struct acopm_dnsbl_request *const request = (struct acopm_dnsbl_request *) data;

	assert(request != NULL);
	assert(request->domain != NULL);
	assert(request->lookup != NULL);
	assert(request->lookup->client != NULL);

	struct acopm_dnsbl_domain *const domain = request->domain;
	struct acopm_dnsbl_lookup *const lookup = request->lookup;
	struct acopm_dnsbl_client *const client = lookup->client;
	const char *const address = lookup->address;

	if (code != DNS_ERR_NONE || type != DNS_IPv4_A || ! count || ! result)
	{
		(void) acopm_log_debug("%s: unusable response for %s", __func__, address);
		(void) acopm_dnsbl_request_free(request);
		return;
	}

	const uint32_t *const resp_addrs = (uint32_t *) result;

	for (int i = 0; i < count; i++)
	{
		char address_str[INET_ADDRSTRLEN];

		if (! inet_ntop(AF_INET, &resp_addrs[i], address_str, sizeof address_str))
		{
			(void) acopm_log_error("%s: inet_ntop: %s", __func__, strerror(errno));
			continue;
		}

		(void) acopm_log_debug("%s: received response to %s%s -> %s",
		                       __func__, lookup->address_rev, domain->host, address_str);

		if (! domain->matches)
		{
			const uint32_t resp_addr_ho = ntohl(resp_addrs[i]);

			if (
				((resp_addr_ho & ACOPM_DNSBL_NOMATCH_MASK) != ACOPM_DNSBL_NOMATCH_MASK) ||
				((resp_addr_ho & ~ACOPM_DNSBL_NOMATCH_MASK) > 0xFF)
			) {
				(void) acopm_log_warning("Ignoring invalid response from DNSBL %s for %s: %s",
				                         domain->host, address, address_str);
				continue;
			}

			(void) acopm_log_debug("%s: no match restrictions, considering DNSBL hit", __func__);
			(void) lookup->callback(client, domain->host, address, address_str, NULL);
			(void) acopm_dnsbl_lookup_free(lookup);
			return;
		}

		struct acopm_dnsbl_match *match = NULL;
		HASH_FIND(hh, domain->matches, &resp_addrs[i], sizeof resp_addrs[i], match);
		if (match)
		{
			(void) acopm_log_debug("%s: DNSBL hit", __func__);
			(void) lookup->callback(client, domain->host, address, address_str, match->desc);
			(void) acopm_dnsbl_lookup_free(lookup);
			return;
		}

		(void) acopm_log_debug("%s: response does not satisfy any match restrictions", __func__);
	}

	(void) acopm_dnsbl_request_free(request);
}

struct acopm_dnsbl_request *
acopm_dnsbl_request_new(struct acopm_dnsbl_lookup *const restrict lookup, struct acopm_dnsbl_domain *const restrict domain)
{
	assert(lookup != NULL);
	assert(lookup->client != NULL);
	assert(lookup->callback != NULL);
	assert(domain != NULL);

	struct acopm_dnsbl_client *const client = lookup->client;

	char address[ACOPM_DNSBL_MAXHOST_LEN + 1];
	if (snprintf(address, sizeof address, "%s%s.", lookup->address_rev, domain->host) > ACOPM_DNSBL_MAXHOST_LEN)
	{
		(void) acopm_log_warning("Name too long to look up: %s", address);
		return NULL;
	}

	struct acopm_dnsbl_request *const request = calloc(1, sizeof *request);
	if (! request)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return NULL;
	}

	request->domain = domain;
	request->lookup = lookup;
	request->request = evdns_base_resolve_ipv4(client->ev_base, address, 0, acopm_dnsbl_resolve_callback, request);

	if (! request->request)
	{
		(void) acopm_log_error("%s: evdns_base_resolve_ipv4: %s", __func__, strerror(errno));
		(void) free(request);
		return NULL;
	}

	(void) acopm_log_debug("%s: created request for name %s", __func__, address);
	DL_APPEND(lookup->requests, request);
	return request;
}

void
acopm_dnsbl_request_free(struct acopm_dnsbl_request *const restrict request)
{
	assert(request != NULL);
	assert(request->lookup != NULL);
	assert(request->lookup->client != NULL);

	struct acopm_dnsbl_lookup *const lookup = request->lookup;
	struct acopm_dnsbl_domain *const domain = request->domain;
	struct acopm_dnsbl_client *const client = lookup->client;

	(void) acopm_log_debug("%s: aborting request (%s%s.)", __func__, lookup->address_rev, domain->host);

	if (request->request)
		(void) evdns_cancel_request(client->ev_base, request->request);

	if (lookup->requests)
		DL_DELETE(lookup->requests, request);

	if (! lookup->requests)
		(void) acopm_dnsbl_lookup_free(lookup);

	(void) free(request);
}
