/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <inttypes.h> /* PRIu16 */
#include <regex.h> /* regcomp(), regerror(), regfree(), REG_* */
#include <string.h> /* memset() */

#include "3rdparty/utlist.h" /* LL_APPEND() */
#include "irc/client.h" /* decls for own functions */
#include "irc/internal.h" /* decls for own functions (private) */
#include "utils/bufconn.h" /* struct acopm_connection, acopm_conn_*() */
#include "utils/log.h" /* acopm_log_*() */

static inline const char * acopm_attr_inline
acopm_redact(const char *const restrict password)
{
	static char result[ACOPM_IRC_PASS_MAX + 1];
	(void) memset(result, 0x00, sizeof result);

	for (size_t x = 0; password[x] != 0x00 && x < sizeof result; x++)
		result[x] = '*';

	return result;
}

struct acopm_irc_client *
acopm_irc_new(struct event_base *const restrict ev_base, const acopm_irc_connect_cb connect_cb,
              const acopm_irc_command_cb command_cb, const acopm_irc_restart_cb restart_cb)
{
	assert(ev_base != NULL);
	assert(connect_cb != NULL);
	assert(command_cb != NULL);
	assert(restart_cb != NULL);

	struct acopm_irc_client *const client = calloc(1, sizeof *client);

	if (! client)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return NULL;
	}

	client->ev_base = ev_base;
	client->connect_cb = connect_cb;
	client->command_cb = command_cb;
	client->restart_cb = restart_cb;

	(void) acopm_strset_buf(client->config.real, ACOPM_IRC_REALNAME_DEFAULT);

	(void) acopm_log_debug("%s: created client", __func__);
	return client;
}

bool
acopm_irc_add_host(struct acopm_irc_client *const restrict client, const char *const restrict name)
{
	assert(client != NULL);
	assert(name != NULL);

	struct acopm_irc_hostname *const host = calloc(1, sizeof *host);

	if (! host)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return false;
	}

	if (! acopm_strset_buf(host->name, name))
		(void) acopm_log_warning("%s: hostname '%s' too long, truncated", __func__, name);

	(void) acopm_log_debug("%s: adding hostname '%s'", __func__, host->name);
	LL_APPEND(client->config.hosts, host);
	return true;
}

void
acopm_irc_set_port(struct acopm_irc_client *const restrict client, const uint16_t port)
{
	assert(client != NULL);
	assert(port != 0);

	(void) snprintf(client->config.port, sizeof client->config.port, "%" PRIu16, port);
	(void) acopm_log_debug("%s: set port to %s", __func__, client->config.port);
}

bool
acopm_irc_set_bind4(struct acopm_irc_client *const restrict client, const struct sockaddr_in *const restrict addr)
{
	assert(client != NULL);
	assert(addr != NULL);

	if (! (client->config.bind4 = calloc(1, sizeof *addr)))
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return false;
	}

	(void) memcpy(client->config.bind4, addr, sizeof *addr);
	(void) acopm_log_debug("%s: set bind address (IPv4)", __func__);
	return true;
}

bool
acopm_irc_set_bind6(struct acopm_irc_client *const restrict client, const struct sockaddr_in6 *const restrict addr)
{
	assert(client != NULL);
	assert(addr != NULL);

	if (! (client->config.bind6 = calloc(1, sizeof *addr)))
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return false;
	}

	(void) memcpy(client->config.bind6, addr, sizeof *addr);
	(void) acopm_log_debug("%s: set bind address (IPv6)", __func__);
	return true;
}

#ifdef BUILD_WITH_MBEDTLS

void
acopm_irc_set_tls(struct acopm_irc_client *const restrict client, const bool use_tls,
                  const enum acopm_conn_certfp_method certfp_method)
{
	assert(client != NULL);

	if (! (client->config.use_tls = use_tls))
		client->config.certfp_method = ACOPM_CONN_CERTFP_METHOD_NONE;
	else
		client->config.certfp_method = certfp_method;
}

bool
acopm_irc_add_tls_certfp(struct acopm_irc_client *const restrict client, const char *const restrict value)
{
	assert(client != NULL);
	assert(value != NULL);

	struct acopm_conn_certfp_value *const certfp_new = calloc(1, sizeof *certfp_new);

	if (! certfp_new)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return false;
	}

	if (! acopm_strset_buf(certfp_new->value, value))
		(void) acopm_log_warning("%s: certfp value '%s' too long, truncated", __func__, value);

	(void) acopm_log_debug("%s: adding certfp value '%s'", __func__, certfp_new->value);
	LL_APPEND(client->config.certfp_values, certfp_new);
	return true;
}

void
acopm_irc_set_tls_cert(struct acopm_irc_client *const restrict client,
                       const char *const restrict cc_file, const char *const restrict ck_file)
{
	assert(client != NULL);
	assert(cc_file != NULL);
	assert(ck_file != NULL);

	if (! acopm_strset_buf(client->config.cc_file, cc_file))
		(void) acopm_log_warning("%s: certificate path '%s' too long, truncated", __func__, cc_file);

	if (! acopm_strset_buf(client->config.ck_file, ck_file))
		(void) acopm_log_warning("%s: private key path '%s' too long, truncated", __func__, ck_file);

	(void) acopm_log_debug("%s: set certificate file to '%s'", __func__, client->config.cc_file);
	(void) acopm_log_debug("%s: set private key file to '%s'", __func__, client->config.ck_file);
}

#endif /* BUILD_WITH_MBEDTLS */

void
acopm_irc_set_nick(struct acopm_irc_client *const restrict client, const char *const restrict nick)
{
	assert(client != NULL);
	assert(nick != NULL);

	if (! acopm_strset_buf(client->config.nick, nick))
		(void) acopm_log_warning("%s: nickname '%s' too long, truncated", __func__, nick);

	(void) acopm_log_debug("%s: set nickname to '%s'", __func__, client->config.nick);
}

void
acopm_irc_set_user(struct acopm_irc_client *const restrict client, const char *const restrict user)
{
	assert(client != NULL);
	assert(user != NULL);

	if (! acopm_strset_buf(client->config.user, user))
		(void) acopm_log_warning("%s: username '%s' too long, truncated", __func__, user);

	(void) acopm_log_debug("%s: set username to '%s'", __func__, client->config.user);
}

void
acopm_irc_set_pass(struct acopm_irc_client *const restrict client, const char *const restrict pass)
{
	assert(client != NULL);
	assert(pass != NULL);

	if (! acopm_strset_buf(client->config.pass, pass))
		(void) acopm_log_warning("%s: password '%s' too long, truncated", __func__, acopm_redact(pass));

	(void) acopm_log_debug("%s: set password to '%s'", __func__, acopm_redact(client->config.pass));
}

void
acopm_irc_set_real(struct acopm_irc_client *const restrict client, const char *const restrict real)
{
	assert(client != NULL);
	assert(real != NULL);

	if (! acopm_strset_buf(client->config.real, real))
		(void) acopm_log_warning("%s: realname '%s' too long, truncated", __func__, real);

	(void) acopm_log_debug("%s: set realname to '%s'", __func__, client->config.real);
}

void
acopm_irc_set_acct(struct acopm_irc_client *const restrict client,
                   const char *const restrict user, const char *const restrict pass)
{
	assert(client != NULL);

	if (user && *user)
	{
		if (! acopm_strset_buf(client->config.acctuser, user))
			(void) acopm_log_warning("%s: username '%s' too long, truncated", __func__, user);

		(void) acopm_log_debug("%s: set account username to '%s'", __func__, client->config.acctuser);
	}

	if (pass && *pass)
	{
		if (! acopm_strset_buf(client->config.acctpass, pass))
			(void) acopm_log_warning("%s: password '%s' too long, truncated", __func__, acopm_redact(pass));

		(void) acopm_log_debug("%s: set account password to '%s'", __func__, acopm_redact(client->config.acctpass));
	}
}

void
acopm_irc_set_oper(struct acopm_irc_client *const restrict client,
                   const char *const restrict user, const char *const restrict pass)
{
	assert(client != NULL);
	assert(user != NULL);
	assert(pass != NULL);

	if (! acopm_strset_buf(client->config.operuser, user))
		(void) acopm_log_warning("%s: opername '%s' too long, truncated", __func__, user);

	if (! acopm_strset_buf(client->config.operpass, pass))
		(void) acopm_log_warning("%s: password '%s' too long, truncated", __func__, acopm_redact(pass));

	(void) acopm_log_debug("%s: set oper user to '%s'", __func__, client->config.operuser);
	(void) acopm_log_debug("%s: set oper pass to '%s'", __func__, acopm_redact(client->config.operpass));
}

void
acopm_irc_set_pcmd(struct acopm_irc_client *const restrict client, const char *const restrict pcmd)
{
	assert(client != NULL);
	assert(pcmd != NULL);

	if (! acopm_strset_buf(client->config.pcmd, pcmd))
		(void) acopm_log_warning("%s: command '%s' too long, truncated", __func__, pcmd);

	(void) acopm_log_debug("%s: set command to '%s'", __func__, client->config.pcmd);
}

void
acopm_irc_set_away(struct acopm_irc_client *const restrict client, const char *const restrict away)
{
	assert(client != NULL);
	assert(away != NULL);

	if (! acopm_strset_buf(client->config.away, away))
		(void) acopm_log_warning("%s: away string '%s' too long, truncated", __func__, away);

	(void) acopm_log_debug("%s: set away to '%s'", __func__, client->config.away);
}

void
acopm_irc_set_mode(struct acopm_irc_client *const restrict client, const char *const restrict mode)
{
	assert(client != NULL);
	assert(mode != NULL);

	if (! acopm_strset_buf(client->config.mode, mode))
		(void) acopm_log_warning("%s: mode string '%s' too long, truncated", __func__, mode);

	(void) acopm_log_debug("%s: set mode to '%s'", __func__, client->config.mode);
}

void
acopm_irc_set_chan(struct acopm_irc_client *const restrict client, const char *const restrict chan,
                   const char *const restrict pass, const unsigned int delay)
{
	assert(client != NULL);

	if (chan)
	{
		if (! acopm_strset_buf(client->config.log_chan, chan))
			(void) acopm_log_warning("%s: channel '%s' too long, truncated", __func__, chan);

		(void) acopm_log_debug("%s: set channel name to '%s'", __func__, client->config.log_chan);
	}

	if (pass)
	{
		if (! acopm_strset_buf(client->config.log_pass, pass))
			(void) acopm_log_warning("%s: key '%s' too long, truncated", __func__, acopm_redact(pass));

		(void) acopm_log_debug("%s: set channel key to '%s'", __func__, acopm_redact(client->config.log_pass));
	}

	if ((client->config.log_delay = delay))
		(void) acopm_log_debug("%s: set channel join delay to %u", __func__, client->config.log_delay);
}

bool
acopm_irc_set_conn(struct acopm_irc_client *const restrict client, const char *const restrict format)
{
	assert(client != NULL);
	assert(format != NULL);

	if (! client->config.conn_regexp)
	{
		if (! acopm_strset_buf(client->config.conn_fmt, format))
			(void) acopm_log_warning("%s: format string '%s' too long, truncated", __func__, format);

		(void) acopm_log_debug("%s: set connection format to '%s'", __func__, client->config.conn_fmt);
		return true;
	}

	(void) memset(&client->config.conn_re, 0x00, sizeof client->config.conn_re);

	const int ret = regcomp(&client->config.conn_re, format, REG_EXTENDED);

	if (ret != 0)
	{
		char errbuf[1024];
		(void) regerror(ret, &client->config.conn_re, errbuf, sizeof errbuf);
		(void) acopm_log_error("%s: regcomp: %s", __func__, errbuf);
		return false;
	}
	else if (client->config.conn_re.re_nsub != ACOPM_IRC_REGEX_PSUBN)
	{
		(void) acopm_log_error("%s: the connection format '%s' does not contain %u parenthetical "
		                       "subexpressions", __func__, format, ACOPM_IRC_REGEX_PSUBN);
		return false;
	}

	return true;
}

void
acopm_irc_set_conn_is_regexp(struct acopm_irc_client *const restrict client, const bool is_regexp)
{
	assert(client != NULL);

	client->config.conn_regexp = is_regexp;
}

void
acopm_irc_free(struct acopm_irc_client *const restrict client)
{
	assert(client != NULL);

	(void) acopm_log_debug("%s: freeing client", __func__);

	(void) acopm_irc_disconnect(client);

	if (client->config.conn_regexp)
		(void) regfree(&client->config.conn_re);

	(void) free(client);
}
