/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <regex.h> /* regexec(), regerror(), REG_* */
#include <string.h> /* memset() */
#include <strings.h> /* strcasecmp() */
#include <unistd.h> /* sleep() */

#include "irc/commands.h" /* decls for own functions (private) */
#include "irc/internal.h" /* decls for own functions (private) */
#include "utils/addr.h" /* acopm_addrisvalid() */
#include "utils/base64.h" /* acopm_base64encode() */
#include "utils/log.h" /* acopm_log_*() */
#include "utils/misc.h" /* acopm_rng_str() */

struct acopm_irc_command
{
	char    *argv[ACOPM_IRC_CMD_SZ];
	size_t   argc;
};

static int
acopm_irccasecmp(const enum acopm_case_mapping case_mapping, const char *const str1, const char *const str2)
{
	assert(str1 != NULL);
	assert(str2 != NULL);

	static const unsigned char acopm_rfc1459_utable[256] = {
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
		0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
		0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
		0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
		0x60, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
		0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x7E, 0x7F,
		0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
		0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
		0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,
		0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,
		0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
		0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
		0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
		0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF,
	};

	static const unsigned char acopm_rfc2812_utable[256] = {
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F,
		0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
		0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
		0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
		0x60, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
		0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x7F,
		0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8A, 0x8B, 0x8C, 0x8D, 0x8E, 0x8F,
		0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
		0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF,
		0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF,
		0xC0, 0xC1, 0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD, 0xCE, 0xCF,
		0xD0, 0xD1, 0xD2, 0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE, 0xDF,
		0xE0, 0xE1, 0xE2, 0xE3, 0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xEB, 0xEC, 0xED, 0xEE, 0xEF,
		0xF0, 0xF1, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF,
	};

	const unsigned char *acopm_irc_utable = NULL;

	switch (case_mapping)
	{
		case ACOPM_CASEMAP_DEFAULT:
		case ACOPM_CASEMAP_RFC1459:
			acopm_irc_utable = acopm_rfc1459_utable;
			break;
		case ACOPM_CASEMAP_RFC2812:
			acopm_irc_utable = acopm_rfc2812_utable;
			break;
		case ACOPM_CASEMAP_ASCII:
			return strcasecmp(str1, str2);
	}

	assert(acopm_irc_utable != NULL);

	for (size_t x = 0; ; x++)
	{
		const unsigned char chr1 = (unsigned char) str1[x];
		const unsigned char chr2 = (unsigned char) str2[x];

		if (chr1 == 0x00 && chr2 == 0x00)
			return 0;

		if (acopm_irc_utable[chr1] == acopm_irc_utable[chr2])
			continue;

		return ((int) (acopm_irc_utable[chr1] - acopm_irc_utable[chr2]));
	}
}

static bool
acopm_irc_cap_done(struct acopm_irc_client *const restrict client)
{
	if (client->state.sent_cap_end)
		return true;

	client->state.sent_cap_end = true;
	(void) acopm_log_info("Ending IRCv3 capability negotiation");
	return acopm_irc_writev(client, "CAP END");
}

static bool
acopm_irc_cap_req_sasl(struct acopm_irc_client *const restrict client,
                       char *const restrict cap_name, char *const restrict cap_value)
{
	if (cap_value)
	{
		char *const end = cap_value + strlen(cap_value);

		char *idx = cap_value;
		while ((idx = strchr(idx, ',')))
			*idx++ = 0x00;

		const char *const desired = ((! *client->config.acctpass) ? "EXTERNAL" : "PLAIN");

		for (char *method = cap_value; method < end; method += strlen(method))
		{
			while (method < end && *method == 0x00)
				method++;

			if (method >= end)
				break;

			if (strcasecmp(method, desired) == 0)
			{
				(void) acopm_log_info("Server supports SASL %s", desired);
				return acopm_irc_writev(client, "CAP REQ :%s", cap_name);
			}
		}

		(void) acopm_log_warning("Server does not support our SASL method, continuing");
		return acopm_irc_cap_done(client);
	}

	(void) acopm_log_info("Server supports SASL, requesting (EXTERNAL/PLAIN support unknown)");
	return acopm_irc_writev(client, "CAP REQ :%s", cap_name);
}

static bool
acopm_irc_cap_parse(struct acopm_irc_client *const restrict client, char *const restrict caps)
{
	char *const end = caps + strlen(caps);

	char *idx = caps;
	while ((idx = strchr(idx, ' ')))
		*idx++ = 0x00;

	char *cap = caps;
	for (;;)
	{
		while (cap < end && *cap == 0x00)
			cap++;

		if (cap >= end)
			break;

		const size_t cap_len = strlen(cap);

		char *value = strchr(cap, '=');
		if (value)
			*value++ = 0x00;

		if (strcasecmp(cap, "SASL") == 0 && *client->config.acctuser)
			return acopm_irc_cap_req_sasl(client, cap, value);

		cap += cap_len;
	}

	(void) acopm_log_warning("Server does not support SASL");
	return acopm_irc_cap_done(client);
}

static bool
acopm_irc_parse_feature(struct acopm_irc_client *const restrict client, char *const restrict feature)
{
	char *value = strchr(feature, '=');

	if (value)
		*value++ = 0x00;

	if (strcasecmp(feature, "CALLERID") == 0 && value && value[0] && ! value[1])
	{
		(void) acopm_log_info("Server supports '%s' (+%s), enabling", feature, value);
		return acopm_irc_writev(client, "MODE %s :+%s", client->state.nick, value);
	}
	if (strcasecmp(feature, "CASEMAPPING") == 0 && value && *value)
	{
		if (strcasecmp(value, "ASCII") == 0)
		{
			(void) acopm_log_info("Server uses ASCII case mapping");
			client->state.case_mapping = ACOPM_CASEMAP_ASCII;
			return true;
		}
		if (strcasecmp(value, "RFC1459") == 0)
		{
			(void) acopm_log_info("Server uses RFC1459 case mapping");
			client->state.case_mapping = ACOPM_CASEMAP_RFC1459;
			return true;
		}
		if (strcasecmp(value, "RFC2812") == 0)
		{
			(void) acopm_log_info("Server uses RFC2812 case mapping");
			client->state.case_mapping = ACOPM_CASEMAP_RFC2812;
			return true;
		}

		(void) acopm_log_warning("Server uses %s case mapping (unrecognised)", value);
		return true;
	}
	if (strcasecmp(feature, "CHANTYPES") == 0 && value && *value)
	{
		if (! acopm_strset_buf(client->state.chantypes, value))
		{
			(void) memset(client->state.chantypes, 0x00, sizeof client->state.chantypes);

			(void) acopm_log_warning("%s: IRCd CHANTYPES value '%s' is longer than limit of %u, ignoring",
			                         __func__, value, ACOPM_IRC_CHANTYPES_MAX);
		}

		return true;
	}
	if (strcasecmp(feature, "STATUSMSG") == 0 && value && *value)
	{
		if (! acopm_strset_buf(client->state.statusmsg, value))
			(void) acopm_log_warning("%s: IRCd STATUSMSG value '%s' is longer than limit of %u, truncated",
			                         __func__, value, ACOPM_IRC_STATUSMSG_MAX);

		return true;
	}

	return true;
}

static bool
acopm_irc_cmd_auth(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (! (buf->argc == 2 && *buf->argv[1] == '+'))
	{
		(void) acopm_log_warning("Received malformed AUTHENTICATE message from server");
		return acopm_irc_cap_done(client);
	}

	const char *const acctuser = client->config.acctuser;
	const char *const acctpass = client->config.acctpass;

	char params[((ACOPM_IRC_NICK_MAX * 2) + ACOPM_IRC_PASS_MAX) + 4];
	char params_b64[(sizeof params * 2) + 4];

	const int ret = snprintf(params, sizeof params, "%s%c%s%c%s", acctuser, 0x00, acctuser, 0x00, acctpass);

	if (ret < 1)
		return false;

	if (((size_t) ret) >= sizeof params)
	{
		(void) acopm_log_critical("%s: account information buffer would overflow (BUG)", __func__);
		return false;
	}

	(void) acopm_base64encode((unsigned char *) params, (size_t) ret, params_b64);

	const char *params_ptr = params_b64;

	/*
	 * If the Base64-encoded string is 400 characters or greater, we must
	 * send it in chunks of 400 characters; and if the last chunk we sent
	 * was 400 characters, we must send an empty message (+) too.
	 *
	 * http://ircv3.net/specs/extensions/sasl-3.1.html
	 */
	for (size_t len = strlen(params_ptr); len >= ACOPM_IRC_SASL_MAX; len -= ACOPM_IRC_SASL_MAX)
	{
		char tmp[ACOPM_IRC_SASL_MAX + 1];
		(void) acopm_strset_buf(tmp, params_ptr);

		if (! acopm_irc_writev(client, "AUTHENTICATE %s", tmp))
			return false;

		if (len == ACOPM_IRC_SASL_MAX)
			return acopm_irc_writev(client, "AUTHENTICATE +");

		params_ptr += ACOPM_IRC_SASL_MAX;
	}

	return acopm_irc_writev(client, "AUTHENTICATE %s", params_ptr);
}

static bool
acopm_irc_cmd_cap(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 4)
	{
		(void) acopm_log_critical("Server sent malformed %s command", buf->argv[0]);
		return false;
	}

	if (strcasecmp(buf->argv[2], "LS") == 0)
	{
		(void) acopm_log_info("Server supports IRCv3");
		return acopm_irc_cap_parse(client, buf->argv[buf->arge]);
	}

	if (strcasecmp(buf->argv[2], "ACK") == 0 && strcasecmp(buf->argv[3], "SASL") == 0)
	{
		if (! *client->config.acctuser)
		{
			(void) acopm_log_critical("%s: no username (BUG)", __func__);
			return acopm_irc_cap_done(client);
		}

		if (! *client->config.acctpass)
		{
			(void) acopm_log_info("Attempting SASL EXTERNAL authentication");
			return acopm_irc_writev(client, "AUTHENTICATE EXTERNAL");
		}
		else
		{
			(void) acopm_log_info("Attempting SASL PLAIN authentication");
			return acopm_irc_writev(client, "AUTHENTICATE PLAIN");
		}
	}

	if (strcasecmp(buf->argv[2], "NAK") == 0 && strcasecmp(buf->argv[3], "SASL") == 0)
	{
		(void) acopm_log_warning("SASL is unavailable");
		return acopm_irc_cap_done(client);
	}

	return false;
}

static bool
acopm_irc_cmd_error(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	(void) client;

	(void) acopm_log_notice("Error from server: %s", buf->argv[buf->arge]);
	return false;
}

static bool
acopm_irc_cmd_invite(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 3)
	{
		(void) acopm_log_warning("Server sent malformed %s command", buf->argv[0]);
		return true;
	}

	const char *const log_chan = client->config.log_chan;
	const char *const log_pass = client->config.log_pass;

	if (client->state.in_log_chan || ! *log_chan)
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->argv[1]) != 0)
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, log_chan, buf->argv[2]) != 0)
		return true;

	if (! acopm_irc_log(client, "Invited to the log channel by %s", buf->source))
		return false;

	char params[ACOPM_IRC_LINE_MAX + 1];
	(void) snprintf(params, sizeof params, "%s%s%s", log_chan, *log_pass ? " " : "", log_pass);

	return acopm_irc_writev(client, "JOIN %s", params);
}

static bool
acopm_irc_cmd_join(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc != 2)
	{
		(void) acopm_log_warning("Server sent malformed %s command", buf->argv[0]);
		return true;
	}

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->source) != 0)
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, client->config.log_chan, buf->argv[1]) != 0)
	{
		(void) acopm_log_notice("Joined to %s without asking -- parting", buf->argv[1]);
		return acopm_irc_writev(client, "PART :%s", buf->argv[1]);
	}

	(void) acopm_log_notice("Joined the log channel");
	client->state.in_log_chan = true;

	if (! *client->state.chantypes)
		if (! acopm_irc_log(client, "Server sent no usable ISUPPORT CHANTYPES token, commands unavailable"))
			return false;

	if (acopm_irccasecmp(client->state.case_mapping, client->config.nick, client->state.nick) != 0)
		return acopm_irc_writev(client, "NICK %s", client->config.nick);

	return true;
}

static bool
acopm_irc_cmd_kick(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 4)
	{
		(void) acopm_log_warning("Server sent malformed %s command", buf->argv[0]);
		return true;
	}

	const char *const log_chan = client->config.log_chan;

	if (! client->state.in_log_chan || ! *log_chan)
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->argv[2]) != 0)
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, client->config.log_chan, buf->argv[1]) != 0)
		return true;

	client->state.in_log_chan = false;
	const char *const reason = buf->argv[buf->arge];
	return acopm_irc_log(client, "Kicked from log channel by %s (%s)", buf->source, reason);
}

static bool
acopm_irc_cmd_kill(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 2)
	{
		(void) acopm_log_warning("Server sent malformed %s command", buf->argv[0]);
		return true;
	}

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->argv[1]) != 0)
		return true;

	const char *const reason = ((buf->argc == 3) ? buf->argv[2] : "No reason");

	(void) acopm_log_notice("Killed by %s (%s)", buf->source, reason);
	(void) client->restart_cb();
	return false;
}

static bool
acopm_irc_cmd_nick(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 2)
	{
		(void) acopm_log_warning("Server sent malformed %s command", buf->argv[0]);
		return true;
	}

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->source) != 0)
	{
		if (acopm_irccasecmp(client->state.case_mapping, client->config.nick, buf->source) != 0)
			return true;

		return acopm_irc_writev(client, "NICK %s", client->config.nick);
	}

	if (! acopm_strset_buf(client->state.nick, buf->argv[1]))
	{
		(void) acopm_log_error("%s: server sent us a new nickname that is too long", __func__);
		return false;
	}

	(void) acopm_log_info("Our nickname has changed to %s", buf->argv[1]);
	return true;
}

static bool
acopm_irc_cmd_notice(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 3)
	{
		(void) acopm_log_warning("Server sent malformed %s command", buf->argv[0]);
		return true;
	}

	if (! buf->source || ! *buf->source || ! strchr(buf->source, '.'))
		return true;

	const char *const snotice = buf->argv[buf->arge];

	char nick[ACOPM_IRC_NICK_MAX + 1];
	char addr[INET6_ADDRSTRLEN];

	if (client->config.conn_regexp)
	{
		regmatch_t matches[ACOPM_IRC_REGEX_TSUBN];
		const regex_t *const conn_re = &client->config.conn_re;

		const size_t snotice_len = strlen(snotice);
		const int ret = regexec(conn_re, snotice, ACOPM_IRC_REGEX_TSUBN, matches, REG_NOTBOL | REG_NOTEOL);

		if (ret == REG_NOMATCH)
		{
			(void) acopm_log_debug("%s: no match", __func__);
			return true;
		}
		else if (ret != 0)
		{
			char errbuf[1024];
			(void) regerror(ret, conn_re, errbuf, sizeof errbuf);
			(void) acopm_log_error("%s: regexec: %s", __func__, errbuf);
			return true;
		}
		else if (matches[1].rm_so < 0 || matches[1].rm_eo < 0)
		{
			(void) acopm_log_warning("%s: no useful string offsets for nickname", __func__);
			return true;
		}
		else if (matches[2].rm_so < 0 || matches[2].rm_eo < 0)
		{
			(void) acopm_log_warning("%s: no useful string offsets for address", __func__);
			return true;
		}
		else if (matches[1].rm_eo <= matches[1].rm_so)
		{
			(void) acopm_log_warning("%s: offsets for nickname invalid", __func__);
			return true;
		}
		else if (matches[2].rm_eo <= matches[2].rm_so)
		{
			(void) acopm_log_warning("%s: offsets for address invalid", __func__);
			return true;
		}
		else if (matches[2].rm_so <= matches[1].rm_eo)
		{
			(void) acopm_log_warning("%s: start of address is <= end of nickname", __func__);
			return true;
		}
		else if (((size_t) matches[1].rm_eo) >= snotice_len)
		{
			(void) acopm_log_warning("%s: end of nickname is past end of string", __func__);
			return true;
		}
		else if (((size_t) matches[2].rm_eo) >= snotice_len)
		{
			(void) acopm_log_warning("%s: end of address is past end of string", __func__);
			return true;
		}

		(void) memset(nick, 0x00, sizeof nick);
		(void) memset(addr, 0x00, sizeof addr);

		(void) memcpy(nick, snotice + matches[1].rm_so, (size_t) (matches[1].rm_eo - matches[1].rm_so));
		(void) memcpy(addr, snotice + matches[2].rm_so, (size_t) (matches[2].rm_eo - matches[2].rm_so));
	}
	else
	{
#ifdef __GNUC__
#  pragma GCC diagnostic push
#  pragma GCC diagnostic ignored "-Wformat-nonliteral"
#endif
		if (sscanf(snotice, client->config.conn_fmt, nick, addr) != 2)
		{
			(void) acopm_log_debug("%s: no match", __func__);
			return true;
		}
#ifdef __GNUC__
#  pragma GCC diagnostic pop
#endif
	}

	if (strlen(addr) < 7 || strcmp(addr, "0.0.0.0") == 0)
		return true;

	if (! acopm_addrisvalid(addr, NULL))
	{
		(void) acopm_log_warning("Client '%s' matched connect format string but "
		                         "has nonsensical IP address '%s'", nick, addr);
		return true;
	}

	(void) client->connect_cb(client, nick, addr);
	return true;
}

static bool
acopm_irc_cmd_part(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 2)
	{
		(void) acopm_log_warning("Server sent malformed %s command", buf->argv[0]);
		return true;
	}

	if (! client->state.in_log_chan)
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->source) != 0)
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, client->config.log_chan, buf->argv[1]) != 0)
		return true;

	client->state.in_log_chan = false;
	return acopm_irc_writev(client, "QUIT :Malicious Server: Parted from log channel without asking");
}

static bool
acopm_irc_cmd_ping(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 2 || (buf->source && ! strchr(buf->source, '.')))
	{
		(void) acopm_log_critical("Server sent malformed %s command", buf->argv[0]);
		return false;
	}

	return acopm_irc_writev(client, "PONG :%s", buf->argv[buf->arge]);
}

static bool
acopm_irc_cmd_privmsg(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 3 || ! buf->source || ! *buf->source)
	{
		(void) acopm_log_warning("Server sent malformed %s command", buf->argv[0]);
		return true;
	}

#ifdef PACKAGE_STRING
	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->argv[1]) == 0)
		if (strcasecmp(buf->argv[2], "\001VERSION\001") == 0)
			return acopm_irc_writev(client, "NOTICE %s :\001VERSION %s\001", buf->source, PACKAGE_STRING);
#endif

	bool to_channel = false;

	for (const char *prefix = client->state.statusmsg; *prefix != 0x00; prefix++)
	{
		if (*buf->argv[1] == *prefix)
		{
			buf->argv[1]++;
			break;
		}
	}
	for (const char *prefix = client->state.chantypes; *prefix != 0x00; prefix++)
	{
		if (*buf->argv[1] == *prefix)
		{
			to_channel = true;
			break;
		}
	}

	if (! to_channel || ! strchr(buf->argv[buf->arge], ' '))
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, client->config.log_chan, buf->argv[1]) != 0)
		return true;

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->source) == 0)
		return true;

	char *const line = buf->argv[buf->arge];
	char *const end = line + strlen(line);
	char *idx = line;

	while ((idx = strchr(idx, ' ')))
		*idx++ = 0x00;

	const size_t pref_len = strlen(line);
	const size_t nick_len = strlen(client->state.nick);

	if (pref_len != (nick_len + 1) || (line[(pref_len - 1)] != ':' && line[(pref_len - 1)] != ','))
		return true;

	line[(pref_len - 1)] = 0x00;

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, line) != 0)
		return true;

	struct acopm_irc_command args;
	(void) memset(&args, 0x00, sizeof args);

	char *argn = line + strlen(line);

	for (;;)
	{
		while (argn < end && *argn == 0x00)
			argn++;

		if (argn >= end)
			break;

		if (args.argc >= ACOPM_IRC_CMD_SZ)
			return true;

		args.argv[args.argc++] = argn;
		argn += strlen(argn);
	}

	if (args.argc)
		(void) client->command_cb(client, args.argc, args.argv);

	return true;
}

static bool
acopm_irc_cmd_quit(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (acopm_irccasecmp(client->state.case_mapping, client->config.nick, buf->source) == 0)
		if (acopm_irccasecmp(client->state.case_mapping, client->config.nick, client->state.nick) != 0)
			return acopm_irc_writev(client, "NICK %s", client->config.nick);

	if (acopm_irccasecmp(client->state.case_mapping, client->state.nick, buf->source) != 0)
		return true;

	(void) acopm_log_warning("Disconnecting");
	return false;
}

static bool
acopm_irc_num_welcome(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 2)
	{
		(void) acopm_log_critical("Server sent malformed %s numeric", buf->argv[0]);
		return false;
	}

	(void) acopm_log_info("Our nickname is %s", buf->argv[1]);

	if (! acopm_strset_buf(client->state.nick, buf->argv[1]))
	{
		(void) acopm_log_error("%s: server sent us a new nickname that is too long", __func__);
		return false;
	}

	const char *const operuser = client->config.operuser;
	const char *const operpass = client->config.operpass;

	if (client->state.sent_oper_creds || ! (*operuser && *operpass))
		return true;

	client->state.sent_oper_creds = true;
	(void) acopm_log_info("Sending operator credentials");

	return acopm_irc_writev(client, "OPER %s %s", operuser, operpass);
}

static bool
acopm_irc_num_isupport(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc == 3)
	{
		char *features = buf->argv[buf->arge];
		char *const end = features + strlen(features);

		char *idx = features;
		while ((idx = strchr(idx, ' ')))
			*idx++ = 0x00;

		char *feature = features;
		for (;;)
		{
			while (feature < end && *feature == 0x00)
				feature++;

			if (feature >= end)
				break;

			const size_t feature_len = strlen(feature);

			if (feature_len && ! acopm_irc_parse_feature(client, feature))
				return false;

			feature += feature_len;
		}

		if (feature == features)
		{
			(void) acopm_log_critical("Server sent malformed %s numeric", buf->argv[0]);
			return false;
		}

		return true;
	}
	if (buf->argc > 3)
	{
		for (size_t x = 2; x < buf->arge; x++)
			if (! acopm_irc_parse_feature(client, buf->argv[x]))
				return false;

		return true;
	}

	(void) acopm_log_critical("Server sent malformed %s numeric", buf->argv[0]);
	return false;
}

static bool
acopm_irc_num_operator(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	(void) buf;

	const char *const log_chan = client->config.log_chan;
	const char *const log_pass = client->config.log_pass;

	(void) acopm_log_notice("We are now an IRC Operator");

	if (*client->config.pcmd && ! acopm_irc_writev(client, "%s", client->config.pcmd))
	{
		(void) acopm_log_error("Cannot send client post-connect command");
		return false;
	}

	if (*client->config.away && ! acopm_irc_writev(client, "AWAY :%s", client->config.away))
	{
		(void) acopm_log_error("Cannot send client away string");
		return false;
	}

	if (*client->config.mode && ! acopm_irc_writev(client, "MODE %s :%s", client->state.nick, client->config.mode))
	{
		(void) acopm_log_error("Cannot send client mode string");
		return false;
	}

	if (client->state.in_log_chan || ! *log_chan)
		return true;

	if (client->config.log_delay)
	{
		(void) acopm_log_info("Delaying channel join for %u seconds", client->config.log_delay);
		(void) sleep(client->config.log_delay);
	}

	char params[ACOPM_IRC_LINE_MAX + 1];
	(void) snprintf(params, sizeof params, "%s%s%s", log_chan, *log_pass ? " " : "", log_pass);

	(void) acopm_log_info("Joining log channel");
	return acopm_irc_writev(client, "JOIN %s", params);
}

static bool
acopm_irc_num_nick_used(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc < 3)
	{
		(void) acopm_log_critical("Server sent malformed %s numeric", buf->argv[0]);
		return false;
	}

	if (client->state.in_log_chan)
		return true;

	char randstr[ACOPM_IRC_RAND_NICK_LEN + 1];

	if (! acopm_rng_str(randstr, sizeof randstr))
	{
		(void) acopm_log_critical("Cannot generate random nickname");
		return false;
	}

	return acopm_irc_writev(client, "NICK ACOPM-%s", randstr);
}

static bool
acopm_irc_num_sasl_pass(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	(void) client;

	if (buf->argc == 5)
		(void) acopm_log_info("%s", buf->argv[buf->arge]);

	return true;
}

static bool
acopm_irc_num_sasl_done(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	if (buf->argc == 3)
		(void) acopm_log_notice("%s", buf->argv[buf->arge]);

	return acopm_irc_cap_done(client);
}

static bool
acopm_irc_num_oper_fail(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	(void) client;

	(void) acopm_log_error("%s: %s", __func__, buf->argv[buf->arge]);
	return false;
}

static bool
acopm_irc_num_sasl_fail(struct acopm_irc_client *const restrict client, struct acopm_irc_cmd_buf *const restrict buf)
{
	(void) acopm_log_warning("%s", buf->argv[buf->arge]);
	return acopm_irc_cap_done(client);
}

bool
acopm_irc_command_dispatch(struct acopm_irc_client *const restrict client,
                           struct acopm_irc_cmd_buf *const restrict buf)
{
	assert(client != NULL);
	assert(buf != NULL);
	assert(*buf->argv[0] != 0x00);

	static const struct
	{

		const char      *command;
		bool           (*handler)(struct acopm_irc_client *, struct acopm_irc_cmd_buf *);

	} acopm_irc_cmd_handlers[] = {

		{        "NOTICE", acopm_irc_cmd_notice     },
		{          "PING", acopm_irc_cmd_ping       },
		{       "PRIVMSG", acopm_irc_cmd_privmsg    },
		{          "JOIN", acopm_irc_cmd_join       },
		{          "NICK", acopm_irc_cmd_nick       },
		{          "KICK", acopm_irc_cmd_kick       },
		{          "PART", acopm_irc_cmd_part       },
		{          "QUIT", acopm_irc_cmd_quit       },
		{        "INVITE", acopm_irc_cmd_invite     },
		{          "KILL", acopm_irc_cmd_kill       },
		{         "ERROR", acopm_irc_cmd_error      },

		{  "AUTHENTICATE", acopm_irc_cmd_auth       },
		{           "CAP", acopm_irc_cmd_cap        },
		{           "001", acopm_irc_num_welcome    },
		{           "005", acopm_irc_num_isupport   },
		{           "381", acopm_irc_num_operator   },
		{           "433", acopm_irc_num_nick_used  },
		{           "464", acopm_irc_num_oper_fail  },
		{           "491", acopm_irc_num_oper_fail  },
		{           "900", acopm_irc_num_sasl_pass  },
		{           "903", acopm_irc_num_sasl_done  },
		{           "904", acopm_irc_num_sasl_fail  },
		{           "905", acopm_irc_num_sasl_fail  },
		{           "906", acopm_irc_num_sasl_fail  },

	};

	static const size_t handler_count = sizeof acopm_irc_cmd_handlers / sizeof acopm_irc_cmd_handlers[0];

	for (size_t x = 0; x < handler_count; x++)
		if (strcasecmp(acopm_irc_cmd_handlers[x].command, buf->argv[0]) == 0)
			return acopm_irc_cmd_handlers[x].handler(client, buf);

	return true;
}
