/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <netdb.h> /* getaddrinfo(), freeaddrinfo(), struct addrinfo */
#include <string.h> /* memset(), strlen() */

#include <netinet/in.h> /* IPPROTO_TCP */
#include <sys/socket.h> /* AF_UNSPEC, SOCK_STREAM */

#include <event2/event.h> /* event_base */

#include "3rdparty/utlist.h" /* LL_FOREACH(), LL_FOREACH_SAFE(), LL_DELETE() */
#include "irc/client.h" /* decls for own functions */
#include "irc/commands.h" /* decls for own functions and structs (private) */
#include "irc/internal.h" /* decls for own functions and structs (private) */
#include "utils/addr.h" /* ACOPM_ADDRSTRLEN */
#include "utils/bufconn.h" /* acopm_conn_*, struct acopm_connection, acopm_conn_*() */
#include "utils/ignore_errno.h" /* acopm_ignore_errno() */
#include "utils/log.h" /* acopm_log_*() */

static inline bool acopm_attr_inline
acopm_irc_parse_buf_add(struct acopm_irc_cmd_buf *const restrict buf, char *const restrict param)
{
	if (buf->argc >= ACOPM_IRC_CMD_SZ)
	{
		(void) acopm_log_critical("%s: argument count has reached maximum %u", __func__, ACOPM_IRC_CMD_SZ);
		return false;
	}

	buf->argv[buf->argc] = param;
	buf->arge = buf->argc++;
	return true;
}

static bool
acopm_irc_parse_line(struct acopm_irc_client *const restrict client, char *restrict line)
{
	(void) acopm_log_traffic("<<< %s", line);

	struct acopm_irc_cmd_buf buf;
	(void) memset(&buf, 0x00, sizeof buf);

	char *arge = strstr(line, " :");
	if (arge)
	{
		*arge++ = 0x00;
		*arge++ = 0x00;
	}

	char *const end = line + strlen(line);

	char *idx = line;
	while ((idx = strchr(idx, ' ')))
		*idx++ = 0x00;

	if (*line == ':')
	{
		*line++ = 0x00;
		const size_t source_len = strlen(line);

		if (source_len)
		{
			buf.source = line;
			line += source_len;

			if ((idx = strchr(buf.source, '!')))
				while (*idx != 0x00)
					*idx++ = 0x00;

			if (! *buf.source)
				buf.source = NULL;
		}
	}

	char *argn = line;
	for (;;)
	{
		while (argn < end && *argn == 0x00)
			argn++;

		if (argn >= end)
			break;

		if (! acopm_irc_parse_buf_add(&buf, argn))
			return false;

		argn += strlen(argn);
	}

	if (! buf.argc)
	{
		(void) acopm_log_critical("%s: no command in line from server", __func__);
		return false;
	}

	if (arge && ! acopm_irc_parse_buf_add(&buf, arge))
		return false;

	return acopm_irc_command_dispatch(client, &buf);
}

static bool
acopm_irc_ev_recv_func(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_irc_client *const client = (struct acopm_irc_client *) connection->context;

	assert(client->state.connection == connection);

	unsigned char *const buf = connection->recvbuf + connection->recvbuf_used;
	const size_t prev = connection->recvbuf_used;

	if (! acopm_conn_readbuf(connection, acopm_irc_ev_recv_func))
		return false;

	const size_t len = connection->recvbuf_used - prev;

	if (memchr(buf, '\0', len))
	{
		(void) acopm_log_critical("%s: received NULL byte from server!", __func__);
		return false;
	}

	unsigned char *idx = NULL;
	char *end = NULL;
	while ((idx = memchr(connection->recvbuf, '\n', connection->recvbuf_used)))
	{
		*idx++ = 0x00;
		end = (char *) idx;
	}
	if (! end)
	{
		if (connection->recvbuf_used >= (ACOPM_IRC_LINE_MAX + 2))
		{
			(void) acopm_log_critical("%s: received line with excessive length from server", __func__);
			return false;
		}

		(void) acopm_log_debug("%s: complete line not found in buffer yet", __func__);
		return true;
	}

	char *line = (char *) connection->recvbuf;
	for (;;)
	{
		while (line < end && *line == 0x00)
			line++;

		if (line >= end)
			break;

                char *tmp;
		while ((tmp = strchr(line, '\r')))
			*tmp++ = 0x00;

		const size_t line_len = strlen(line);

		if (line_len > ACOPM_IRC_LINE_MAX)
		{
			(void) acopm_log_critical("%s: server sent line of length %zu", __func__, line_len);
			return false;
		}

		if (line_len && ! acopm_irc_parse_line(client, line))
			return false;

		line += line_len;
	}

	const size_t tmp = (size_t) (((unsigned char *) end) - connection->recvbuf);
	const size_t rem = connection->recvbuf_used - tmp;

	if ((connection->recvbuf_used = rem))
	{
		(void) acopm_log_debug("%s: moving %zu bytes from %zu to start of recv buffer", __func__, rem, tmp);

		(void) memmove(connection->recvbuf, end, rem);
	}

	(void) acopm_log_debug("%s: now have %zu/%zu bytes in recv buffer",
	                       __func__, connection->recvbuf_used, connection->recvbuf_size);
	return true;
}

static bool
acopm_irc_ev_reg_func(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_irc_client *const client = (struct acopm_irc_client *) connection->context;

	assert(client->state.connection == connection);

	const bool use_ircv3 = ((*client->config.acctuser) ? true : false);

	(void) acopm_log_info("Registering with server (IRCv3 %sabled)", ((use_ircv3) ? "En" : "Dis"));

	if (use_ircv3 && ! acopm_irc_writev(client, "CAP LS 302"))
		return false;

	if (*client->config.pass && ! acopm_irc_writev(client, "PASS %s", client->config.pass))
		return false;

	if (! acopm_irc_writev(client, "USER %s 8 * :%s", client->config.user, client->config.real))
		return false;

	if (! acopm_irc_writev(client, "NICK %s", client->config.nick))
		return false;

	return acopm_conn_ev_mod(connection, EV_READ, EV_WRITE, acopm_irc_ev_recv_func, NULL);
}

static inline void acopm_attr_inline
acopm_irc_clear_bind(struct acopm_irc_client *const restrict client)
{
	assert(client != NULL);

	if (client->config.bind4)
	{
		(void) free(client->config.bind4);
		client->config.bind4 = NULL;
	}
	if (client->config.bind6)
	{
		(void) free(client->config.bind6);
		client->config.bind6 = NULL;
	}
}

static inline void acopm_attr_inline
acopm_irc_clear_hosts(struct acopm_irc_client *const restrict client)
{
	assert(client != NULL);

	struct acopm_irc_hostname *cur, *tmp;
	LL_FOREACH_SAFE(client->config.hosts, cur, tmp)
	{
		LL_DELETE(client->config.hosts, cur);
		(void) free(cur);
	}
}

#ifdef BUILD_WITH_MBEDTLS

static inline void acopm_attr_inline
acopm_irc_clear_certfp(struct acopm_irc_client *const restrict client)
{
	assert(client != NULL);

	struct acopm_conn_certfp_value *cur, *tmp;
	LL_FOREACH_SAFE(client->config.certfp_values, cur, tmp)
	{
		LL_DELETE(client->config.certfp_values, cur);
		(void) free(cur);
	}
}

#endif

static bool
acopm_irc_connect_each(struct acopm_irc_client *const restrict client, struct addrinfo *const restrict addr)
{
	assert(client != NULL);
	assert(addr != NULL);
	assert(addr->ai_addr != NULL);
	assert(addr->ai_addr->sa_family == AF_INET || addr->ai_addr->sa_family == AF_INET6);

	struct acopm_connection *const connection = acopm_conn_new("IRC/Client", client);

	if (! (client->state.connection = connection))
		return false;

	if (addr->ai_addr->sa_family == AF_INET)
		(void) acopm_conn_set_bind4(connection, client->config.bind4);
	else if (addr->ai_addr->sa_family == AF_INET6)
		(void) acopm_conn_set_bind6(connection, client->config.bind6);

	const acopm_conn_data_func ev_rfs = NULL;
	const acopm_conn_data_func ev_sfs = acopm_irc_ev_reg_func;
	const short ev_es = EV_WRITE;

	(void) acopm_conn_set_data_cb(connection, ev_es, ev_rfs, ev_sfs);
	(void) acopm_conn_set_ev_break(connection, true);
	(void) acopm_conn_set_nonblock(connection, false, 0);

#ifdef BUILD_WITH_MBEDTLS
	const bool use_tls = client->config.use_tls;
	const enum acopm_conn_certfp_method certfp_method = client->config.certfp_method;
	const struct acopm_conn_certfp_value *const certfp_values = client->config.certfp_values;
	const char *const cc_file = client->config.cc_file;
	const char *const ck_file = client->config.ck_file;

	if (! acopm_conn_set_tls(connection, use_tls, certfp_method, certfp_values))
		return false;

	if (use_tls && *cc_file && *ck_file)
		(void) acopm_conn_set_tls_cert(connection, cc_file, ck_file);
#endif

	if (! acopm_conn_do_connect(connection, client->ev_base, addr->ai_addr, addr->ai_addrlen))
		return false;

	char al[ACOPM_ADDRSTRLEN];
	char ar[ACOPM_ADDRSTRLEN];
	if (acopm_conn_get_sockaddr(connection, al, ar))
		(void) acopm_log_info("Connected to IRCd (%s -> %s)", al, ar);

	(void) acopm_irc_clear_bind(client);
	(void) acopm_irc_clear_hosts(client);

	return true;
}

static bool
acopm_irc_connect_host(struct acopm_irc_client *const restrict client, const char *const restrict host)
{
	assert(client != NULL);

	(void) acopm_log_notice("Looking up hostname '%s' ...", host);

	struct addrinfo hints;
	(void) memset(&hints, 0x00, sizeof hints);

	hints.ai_flags = AI_NUMERICSERV;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;

	struct addrinfo *gai_list = NULL;
	const int gai_errno = getaddrinfo(host, client->config.port, &hints, &gai_list);

	if (gai_errno != 0)
	{
		(void) acopm_log_error("%s: getaddrinfo: %s", __func__, gai_strerror(gai_errno));
		return false;
	}
	if (! gai_list)
	{
		(void) acopm_log_error("%s: getaddrinfo: no addresses", __func__);
		return false;
	}

	(void) acopm_log_notice("Connecting to server ...");

	for (struct addrinfo *gai_addr = gai_list; gai_addr != NULL; gai_addr = gai_addr->ai_next)
	{
		if (acopm_irc_connect_each(client, gai_addr))
		{
			(void) freeaddrinfo(gai_list);
			return true;
		}
		else
			(void) acopm_irc_disconnect(client);
	}

	(void) acopm_log_warning("%s: connection to %s failed: exhausted all possible addresses", __func__, host);
	(void) freeaddrinfo(gai_list);
	return false;
}

bool
acopm_irc_connect(struct acopm_irc_client *const restrict client)
{
	assert(client != NULL);

	if (! client->config.hosts || ! *client->config.port)
	{
		(void) acopm_log_error("Cannot connect to server: no hosts or port set for client");
		return false;
	}
	if (! *client->config.user)
	{
		(void) acopm_log_error("Cannot connect to server: no username set for client");
		return false;
	}
	if (! *client->config.nick)
	{
		(void) acopm_log_error("Cannot connect to server: no nickname set for client");
		return false;
	}
	if (! *client->config.operuser || ! *client->config.operpass)
	{
		(void) acopm_log_error("Cannot connect to server: no operator set for client");
		return false;
	}

	if (! *client->config.conn_fmt)
	{
		(void) acopm_log_error("Cannot connect to server: no connection format string set for client");
		return false;
	}
	else if (! client->config.conn_regexp && strstr(client->config.conn_fmt, ".+"))
		(void) acopm_log_warning("Warning: options::conn_regexp is false, but server::conn_fmt appears to "
		                         "be a regular expression -- continuing anyway");
	else if (client->config.conn_regexp && strstr(client->config.conn_fmt, "%s"))
		(void) acopm_log_warning("Warning: options::conn_regexp is true, but server::conn_fmt appears to "
		                         "be a sscanf(3) format string -- continuing anyway");

	struct acopm_irc_hostname *host;
	LL_FOREACH(client->config.hosts, host)
		if (acopm_irc_connect_host(client, host->name))
			return true;

	(void) acopm_log_error("%s: connection to server failed: exhausted all possible addresses", __func__);
	return false;
}

bool
acopm_irc_writev(struct acopm_irc_client *const restrict client, const char *const restrict format, ...)
{
	assert(client != NULL);
	assert(client->state.connection != NULL);

	struct acopm_connection *const connection = client->state.connection;

	char message[ACOPM_IRC_LINE_MAX + 3];
	va_list argv;

	va_start(argv, format);
	const int fret = vsnprintf(message, ACOPM_IRC_LINE_MAX, format, argv);
	va_end(argv);

	if (fret < 1)
	{
		(void) acopm_log_critical("%s: vsnprintf(3) failed", __func__);
		return false;
	}
	if (((size_t) fret) >= ACOPM_IRC_LINE_MAX)
	{
		(void) acopm_log_critical("%s: line too long, not sending '%s'", __func__, message);
		return false;
	}

	(void) acopm_log_traffic(">>> %s", message);

	const size_t message_length = ((size_t) fret) + 2;
	message[message_length - 2] = '\r';
	message[message_length - 1] = '\n';

	return acopm_conn_writebuf(connection, (void *) message, message_length);
}

bool
acopm_irc_log(struct acopm_irc_client *const restrict client, const char *const restrict format, ...)
{
	assert(client != NULL);

	char message[ACOPM_IRC_LINE_MAX + 1];
	va_list argv;

	va_start(argv, format);
	const int fret = vsnprintf(message, ACOPM_IRC_LINE_MAX, format, argv);
	va_end(argv);

	if (fret < 1)
	{
		(void) acopm_log_critical("%s: vsnprintf(3) failed", __func__);
		return false;
	}
	if (((size_t) fret) >= ACOPM_IRC_LINE_MAX)
	{
		(void) acopm_log_critical("%s: message too long, not sending '%s'", __func__, message);
		return false;
	}

	(void) acopm_log_notice("%s", message);

	if (client->state.in_log_chan)
		return acopm_irc_writev(client, "PRIVMSG %s :%s", client->config.log_chan, message);
	else
		return acopm_irc_writev(client, "WALLOPS :%s", message);
}

void
acopm_irc_disconnect(struct acopm_irc_client *const restrict client)
{
	assert(client != NULL);

	if (client->state.connection)
	{
		(void) acopm_conn_free(client->state.connection);
		client->state.connection = NULL;
	}

#ifdef BUILD_WITH_MBEDTLS
	(void) acopm_irc_clear_certfp(client);
#endif

	(void) acopm_irc_clear_bind(client);
	(void) acopm_irc_clear_hosts(client);
}
