/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INC_LIBACOPM_IRC_INTERNAL_H
#define INC_LIBACOPM_IRC_INTERNAL_H

#include "acopm-common.h"

#include <limits.h> /* PATH_MAX */
#include <regex.h> /* regex_t */

#include <event2/event.h> /* struct event_base */

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

#include "utils/bufconn.h" /* acopm_conn_*, struct acopm_connection */
#include "irc/client.h" /* decls for own functions */

#define ACOPM_IRC_REALNAME_DEFAULT      "AlphaChat Open Proxy Monitor"

#define ACOPM_IRC_HOST_MAX              255U
#define ACOPM_IRC_PORT_MAX              5U
#define ACOPM_IRC_NICK_MAX              32U
#define ACOPM_IRC_REAL_MAX              150U /* 3x what ratbox et al allow by default */
#define ACOPM_IRC_USER_MAX              10U
#define ACOPM_IRC_PASS_MAX              480U
#define ACOPM_IRC_SASL_MAX              400U /* http://ircv3.net/specs/extensions/sasl-3.1.html */
#define ACOPM_IRC_OPER_MAX              96U
#define ACOPM_IRC_MODE_MAX              112U /* "+-[A-Za-z] +-[A-Za-z]" */
#define ACOPM_IRC_CHAN_MAX              64U
#define ACOPM_IRC_AWAY_MAX              360U
#define ACOPM_IRC_CHANTYPES_MAX         16U  /* #channelname, &channelname, !channelname, ... */
#define ACOPM_IRC_STATUSMSG_MAX         16U  /* PRIVMSG +#channel :foo, PRIVMSG @#channel :foo, ... */

#define ACOPM_IRC_RAND_NICK_LEN         3U
#define ACOPM_IRC_REGEX_PSUBN           2U
#define ACOPM_IRC_REGEX_TSUBN           (ACOPM_IRC_REGEX_PSUBN + 1U)

enum acopm_case_mapping
{
	ACOPM_CASEMAP_DEFAULT           = 0,
	ACOPM_CASEMAP_RFC1459           = 1,
	ACOPM_CASEMAP_RFC2812           = 2,
	ACOPM_CASEMAP_ASCII             = 3,
} acopm_attr_packed;

struct acopm_irc_hostname
{
	struct acopm_irc_hostname      *next;
	char                            name[ACOPM_IRC_HOST_MAX + 1];
} acopm_attr_packed;

struct acopm_irc_client_config
{
	regex_t                         conn_re;

	struct sockaddr_in             *bind4;
	struct sockaddr_in6            *bind6;
	struct acopm_irc_hostname      *hosts;

#ifdef BUILD_WITH_MBEDTLS
	struct acopm_conn_certfp_value *certfp_values;
#endif

	unsigned int                    log_delay;

#ifdef BUILD_WITH_MBEDTLS
	enum acopm_conn_certfp_method   certfp_method;
#endif

	char                            port[ACOPM_IRC_PORT_MAX + 1];
	char                            nick[ACOPM_IRC_NICK_MAX + 1];
	char                            real[ACOPM_IRC_REAL_MAX + 1];
	char                            user[ACOPM_IRC_USER_MAX + 1];
	char                            pass[ACOPM_IRC_PASS_MAX + 1];
	char                            operuser[ACOPM_IRC_OPER_MAX + 1];
	char                            operpass[ACOPM_IRC_PASS_MAX + 1];
	char                            acctuser[ACOPM_IRC_NICK_MAX + 1];
	char                            acctpass[ACOPM_IRC_PASS_MAX + 1];
	char                            pcmd[ACOPM_IRC_LINE_MAX + 1];
	char                            away[ACOPM_IRC_AWAY_MAX + 1];
	char                            mode[ACOPM_IRC_MODE_MAX + 1];
	char                            conn_fmt[ACOPM_IRC_LINE_MAX + 1];
	char                            log_chan[ACOPM_IRC_CHAN_MAX + 1];
	char                            log_pass[ACOPM_IRC_PASS_MAX + 1];

#ifdef BUILD_WITH_MBEDTLS
	char                            cc_file[PATH_MAX + 1];
	char                            ck_file[PATH_MAX + 1];

	bool                            use_tls:1;
#endif

	bool                            conn_regexp:1;
} acopm_attr_packed;

struct acopm_irc_client_state
{
	struct acopm_connection        *connection;

	enum acopm_case_mapping         case_mapping;

	char                            nick[ACOPM_IRC_NICK_MAX + 1];
	char                            chantypes[ACOPM_IRC_CHANTYPES_MAX + 1];
	char                            statusmsg[ACOPM_IRC_STATUSMSG_MAX + 1];

	bool                            sent_cap_end:1;
	bool                            sent_oper_creds:1;
	bool                            in_log_chan:1;
} acopm_attr_packed;

struct acopm_irc_client
{
	struct event_base              *ev_base;

	acopm_irc_connect_cb            connect_cb;
	acopm_irc_command_cb            command_cb;
	acopm_irc_restart_cb            restart_cb;

	struct acopm_irc_client_config  config;
	struct acopm_irc_client_state   state;
} acopm_attr_packed;

#endif /* !INC_LIBACOPM_IRC_INTERNAL_H */
