/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <string.h> /* memset() */

#include <arpa/inet.h> /* inet_pton() */
#include <sys/socket.h> /* AF_INET, AF_INET6 */

#include "3rdparty/utlist.h" /* DL_*() */
#include "proxy/host.h" /* decls for own functions */
#include "proxy/internal.h" /* internal decls */
#include "proxy/protos.h" /* acopm_*_send_*() */
#include "utils/bufconn.h" /* struct acopm_connection, acopm_conn_*() */
#include "utils/log.h" /* acopm_log_*() */

static inline bool acopm_attr_inline
acopm_proxy_decide_connvars(const enum acopm_proxy_type proxy_type, acopm_proxy_setup_func *const restrict setup_func,
                            acopm_conn_data_func *const restrict ev_sfs, bool *const restrict use_tls)
{
	switch (proxy_type)
	{
		case ACOPM_PROXY_TYPE_NONE:
			break;

		case ACOPM_PROXY_TYPE_PORT:
			*setup_func = NULL;
			*ev_sfs = NULL;
			*use_tls = false;
			return true;

		case ACOPM_PROXY_TYPE_SOCKS4:
			*setup_func = NULL;
			*ev_sfs = acopm_socks4_send_request;
			*use_tls = false;
			return true;

		case ACOPM_PROXY_TYPE_SOCKS5:
			*setup_func = NULL;
			*ev_sfs = acopm_socks5_send_methods;
			*use_tls = false;
			return true;

		case ACOPM_PROXY_TYPE_HTTP_CONNECT:
			*setup_func = NULL;
			*ev_sfs = acopm_http_send_request;
			*use_tls = false;
			return true;

#ifdef BUILD_WITH_MBEDTLS
		case ACOPM_PROXY_TYPE_HTTPS_CONNECT:
			*setup_func = NULL;
			*ev_sfs = acopm_http_send_request;
			*use_tls = true;
			return true;
#endif

	}

	(void) acopm_log_critical("%s: unknown proxy type %u (BUG)", __func__, (unsigned int) proxy_type);
	return false;
}

static bool
acopm_proxy_write_callback(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_proxy_scan *const scan = (struct acopm_proxy_scan *) connection->context;

	assert(scan->host != NULL);
	assert(scan->port != NULL);
	assert(scan->conn != NULL);
	struct acopm_proxy_host *const host = scan->host;
	struct acopm_proxy_port *const port = scan->port;
	struct acopm_connection *const conn = scan->conn;

	assert(host->client != NULL);
	assert(port->client != NULL);
	assert(host->client == port->client);
	struct acopm_proxy_client *const client = host->client;

#ifdef BUILD_WITH_DEBUG
	char al[ACOPM_ADDRSTRLEN];
	char ar[ACOPM_ADDRSTRLEN];
	if (acopm_conn_get_sockaddr(conn, al, ar))
		(void) acopm_log_debug("%s(%s): connected (%s -> %s)", __func__, conn->name, al, ar);
#endif

	if (port->type == ACOPM_PROXY_TYPE_PORT)
	{
		(void) host->callback(client, host->address_str, port->port, port->type);
		(void) acopm_proxy_host_free(host);
		return true;
	}

	acopm_proxy_setup_func proxy_setup_func = NULL;
	acopm_conn_data_func ev_rfs = NULL;
	acopm_conn_data_func ev_sfs = NULL;
	bool use_tls = false;

	if (! acopm_proxy_decide_connvars(port->type, &proxy_setup_func, &ev_sfs, &use_tls))
		return false;

	if (proxy_setup_func && ! proxy_setup_func(scan))
		return false;

	return acopm_conn_ev_mod(conn, EV_WRITE, EV_READ, ev_rfs, ev_sfs);
}

static void
acopm_proxy_error_callback(struct acopm_connection *const restrict connection, const enum acopm_conn_error errcode)
{
	assert(connection != NULL);
	assert(connection->context != NULL);
	assert(errcode != ACOPM_CONN_NO_ERROR);

	struct acopm_proxy_scan *const scan = (struct acopm_proxy_scan *) connection->context;

	(void) errcode;
	(void) acopm_proxy_scan_free(scan);
}

static bool
acopm_proxy_host_scan(struct acopm_proxy_host *const restrict host, struct acopm_proxy_port *const restrict port)
{
	assert(host != NULL);
	assert(port != NULL);
	assert(host->client != NULL);
	assert(port->client != NULL);
	assert(host->client == port->client);
	assert(port->port != 0);
	assert(port->type != ACOPM_PROXY_TYPE_NONE);

	assert((host->address_family == AF_INET) || (host->address_family == AF_INET6));
	assert((host->address_family == AF_INET) || (port->type != ACOPM_PROXY_TYPE_SOCKS4));

	struct acopm_proxy_client *const client = host->client;

	acopm_proxy_setup_func proxy_setup_func = NULL;
	acopm_conn_data_func ev_sfs = NULL;
	bool use_tls = false;

	if (! acopm_proxy_decide_connvars(port->type, &proxy_setup_func, &ev_sfs, &use_tls))
		return false;

	(void) proxy_setup_func;

	struct acopm_proxy_scan *const scan = calloc(1, sizeof *scan);

	if (! scan)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return false;
	}

	scan->host = host;
	scan->port = port;

	socklen_t sa_len;
	struct sockaddr_storage ss;
	(void) memset(&ss, 0x00, sizeof ss);

	struct sockaddr *const sa = (struct sockaddr *) &ss;
	struct sockaddr_in *const sa4 = (struct sockaddr_in *) &ss;
	struct sockaddr_in6 *const sa6 = (struct sockaddr_in6 *) &ss;

	switch ((sa->sa_family = (sa_family_t) host->address_family))
	{
		case AF_INET:
			sa_len = sizeof *sa4;
			sa4->sin_port = (unsigned short) htons(port->port);
			(void) memcpy(&sa4->sin_addr, host->address_raw, INET_ADDRLEN);
			break;

		case AF_INET6:
			sa_len = sizeof *sa6;
			sa6->sin6_port = (unsigned short) htons(port->port);
			(void) memcpy(&sa6->sin6_addr, host->address_raw, INET6_ADDRLEN);
			break;

		default:
			(void) acopm_log_critical("%s: unknown family %d (BUG)", __func__, (int) sa->sa_family);
			(void) free(scan);
			return false;
	}

	char conn_desc[256];
	(void) snprintf(conn_desc, sizeof conn_desc, "ProxyScan/%s/%u/%u",
	                host->address_str, (unsigned int) port->port, (unsigned int) port->type);

	if (! (scan->conn = acopm_conn_new(conn_desc, scan)))
	{
		(void) free(scan);
		return false;
	}

	(void) acopm_conn_set_data_cb(scan->conn, EV_WRITE, NULL, acopm_proxy_write_callback);
	(void) acopm_conn_set_error_cb(scan->conn, acopm_proxy_error_callback);
	(void) acopm_conn_set_ev_break(scan->conn, false);
	(void) acopm_conn_set_nonblock(scan->conn, true, client->connect_timeout);
	(void) acopm_conn_set_data_timeout(scan->conn, client->data_timeout);

	const size_t recvbuf_sz = ((client->recvbuf_sz) ? client->recvbuf_sz : ACOPM_PROXY_RECVBUF_SZ_DEF);
	const size_t sendbuf_sz = ((client->sendbuf_sz) ? client->sendbuf_sz : ACOPM_PROXY_SENDBUF_SZ_DEF);

	(void) acopm_conn_set_bufsz(scan->conn, recvbuf_sz, sendbuf_sz);

#ifdef BUILD_WITH_MBEDTLS
	if (! acopm_conn_set_tls(scan->conn, use_tls, ACOPM_CONN_CERTFP_METHOD_NONE, NULL))
	{
		(void) acopm_conn_free(scan->conn);
		(void) free(scan);
		return false;
	}
#endif

	if (! acopm_conn_do_connect(scan->conn, client->ev_base, sa, sa_len))
	{
		(void) acopm_conn_free(scan->conn);
		(void) free(scan);
		return false;
	}

	DL_APPEND(host->scans, scan);
	return true;
}

void
acopm_proxy_scan_free(struct acopm_proxy_scan *const restrict scan)
{
	assert(scan != NULL);
	assert(scan->host != NULL);
	assert(scan->port != NULL);

	struct acopm_proxy_host *const host = scan->host;

	(void) acopm_log_debug("%s: freeing scan %u/%u from host %s", __func__,
	                       (unsigned int) scan->port->port, (unsigned int) scan->port->type, host->address_str);

	if (scan->conn)
		(void) acopm_conn_free(scan->conn);

	if (host->scans)
		DL_DELETE(host->scans, scan);

	if (! host->scans)
		(void) acopm_proxy_host_free(host);

	(void) free(scan);
}

struct acopm_proxy_host *
acopm_proxy_host_new(struct acopm_proxy_client *const restrict client, const char *const restrict address,
                     const acopm_proxy_callback callback)
{
	assert(client != NULL);
	assert(address != NULL);

	if (! client->ports)
		return NULL;

	struct acopm_proxy_host *const host = calloc(1, sizeof *host);

	if (! host)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return NULL;
	}
	if (
		inet_pton((host->address_family = AF_INET), address, host->address_raw) != 1 &&
		inet_pton((host->address_family = AF_INET6), address, host->address_raw) != 1
	) {
		(void) acopm_log_error("%s: inet_pton: %s", __func__, strerror(errno));
		(void) free(host);
		return NULL;
	}

	host->client = client;
	host->callback = callback;

	if (! acopm_strset_buf(host->address_str, address))
	{
		(void) acopm_log_error("%s: address '%s' too long", __func__, address);
		(void) free(host);
		return NULL;
	}

	size_t scanports = 0;
	struct acopm_proxy_port *port;
	DL_FOREACH(client->ports, port)
	{
		if (host->address_family == AF_INET6 && port->type == ACOPM_PROXY_TYPE_SOCKS4)
			continue; // SOCKS4 does not support IPv6

		if (acopm_proxy_host_scan(host, port))
			scanports++;
	}
	if (! scanports)
	{
		(void) acopm_log_error("%s: no ports could be scanned on host %s", __func__, address);
		(void) free(host);
		return NULL;
	}

	(void) acopm_log_debug("%s: created host %s, scanning %zu ports", __func__, address, scanports);
	DL_APPEND(client->hosts, host);
	return host;
}

void
acopm_proxy_host_free(struct acopm_proxy_host *const restrict host)
{
	assert(host != NULL);
	assert(host->client != NULL);

	if (host->scans)
	{
		struct acopm_proxy_scan *scan, *tmp;
		DL_FOREACH_SAFE(host->scans, scan, tmp)
		{
			if (! host->scans)
				break;

			(void) acopm_proxy_scan_free(scan);
		}
	}
	else
	{
		(void) acopm_log_debug("%s: freeing host %s from client", __func__, host->address_str);

		struct acopm_proxy_client *const client = host->client;

		if (client->hosts)
			DL_DELETE(client->hosts, host);

		(void) free(host);
	}
}
