/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <inttypes.h> /* PRIu16 */
#include <string.h> /* strerror() */

#include "3rdparty/utlist.h" /* DL_*() */
#include "proxy/internal.h" /* internal decls */
#include "proxy/port.h" /* decls for own functions */
#include "utils/log.h" /* acopm_log_*() */

struct acopm_proxy_port *
acopm_proxy_port_new(struct acopm_proxy_client *const restrict client)
{
	assert(client != NULL);

	struct acopm_proxy_port *const port = calloc(1, sizeof *port);

	if (! port)
	{
		(void) acopm_log_error("%s: calloc: %s", __func__, strerror(errno));
		return NULL;
	}

	port->client = client;

	(void) acopm_log_debug("%s: created port", __func__);
	return port;
}

void
acopm_proxy_port_set_port(struct acopm_proxy_port *const restrict port, const uint16_t portnum)
{
	assert(port != NULL);
	assert(port->client != NULL);

	(void) acopm_log_debug("%s: setting port number to %" PRIu16, __func__, portnum);

	port->port = portnum;
}

void
acopm_proxy_port_set_type(struct acopm_proxy_port *const restrict port, const enum acopm_proxy_type type)
{
	assert(port != NULL);
	assert(port->client != NULL);
	assert(type != ACOPM_PROXY_TYPE_NONE);

	(void) acopm_log_debug("%s: setting port type to %u", __func__, (unsigned int) type);

	port->type = type;
}

bool
acopm_proxy_port_add(struct acopm_proxy_port *const restrict port)
{
	assert(port != NULL);
	assert(port->client != NULL);

	struct acopm_proxy_client *const client = port->client;

	if (! port->port)
	{
		(void) acopm_log_error("%s: attempt to add port to client without a port number", __func__);
		(void) acopm_proxy_port_free(port);
		return false;
	}
	if (! port->type)
	{
		(void) acopm_log_error("%s: attempt to add port to client without a port type", __func__);
		(void) acopm_proxy_port_free(port);
		return false;
	}

	struct acopm_proxy_port *tmp;
	DL_FOREACH(client->ports, tmp)
	{
		if (tmp->port == port->port && tmp->type == port->type)
		{
			(void) acopm_log_error("%s: attempt to add existing port/type %u/%u to client",
			                       __func__, (unsigned int) port->port, (unsigned int) port->type);

			(void) acopm_proxy_port_free(port);
			return false;
		}
	}

	(void) acopm_log_debug("%s: adding port to client", __func__);
	DL_APPEND(client->ports, port);
	return true;
}

void
acopm_proxy_port_free(struct acopm_proxy_port *const restrict port)
{
	assert(port != NULL);
	assert(port->client != NULL);

	struct acopm_proxy_client *const client = port->client;

	(void) acopm_log_debug("%s: freeing port %u/%u on client", __func__,
	                       (unsigned int) port->port, (unsigned int) port->type);

	if (client->ports)
		DL_DELETE(client->ports, port);

	(void) free(port);
}
