/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include <event2/event.h>

#include "proxy/client.h"
#include "proxy/host.h"
#include "proxy/internal.h"
#include "proxy/port.h"
#include "proxy/protos.h"
#include "utils/bufconn.h"
#include "utils/log.h"

static bool acopm_http_recv_response(struct acopm_connection *);

bool
acopm_http_send_request(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_proxy_scan *const scan = (struct acopm_proxy_scan *) connection->context;

	assert(scan->conn == connection);
	assert(scan->host != NULL);
	assert(scan->port != NULL);
	assert(scan->host->client != NULL);
	assert(scan->port->client != NULL);
	assert(scan->host->client == scan->port->client);

	struct acopm_proxy_host *const host = scan->host;
	struct acopm_proxy_port *const port = scan->port;

	char address[INET6_ADDRSTRLEN + 4];
	char request[1024];

	if (host->address_family == AF_INET)
	{
		if (snprintf(address, sizeof address, "%s", host->address_str) <= 6)
			return false;
	}
	else
	{
		if (snprintf(address, sizeof address, "[%s]", host->address_str) <= 4)
			return false;
	}

	const int ret = snprintf(request, sizeof request, "CONNECT %s:%u HTTP/1.1\r\nHost: %s:%u\r\n\r\n",
	                         address, (unsigned int) port->port, address, (unsigned int) port->port);

	if (ret <= 40)
		return false;

	if (! acopm_conn_writebuf(connection, request, (size_t) ret))
		return false;

	return acopm_conn_ev_mod(connection, EV_READ, EV_WRITE, acopm_http_recv_response, NULL);
}

static bool
acopm_http_recv_response(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_proxy_scan *const scan = (struct acopm_proxy_scan *) connection->context;

	assert(scan->conn == connection);
	assert(scan->host != NULL);
	assert(scan->port != NULL);
	assert(scan->host->client != NULL);
	assert(scan->port->client != NULL);
	assert(scan->host->client == scan->port->client);

	struct acopm_proxy_host *const host = scan->host;
	struct acopm_proxy_port *const port = scan->port;
	struct acopm_proxy_client *const client = host->client;

	if (! acopm_conn_readbuf(connection, acopm_http_recv_response))
		return false;

	if (! connection->recvbuf_used)
		return true;

	const size_t avail = ((connection->recvbuf_used <= 12) ? connection->recvbuf_used : 12);

	switch (avail)
	{
		case 12:
			if (connection->recvbuf[11] < '0' || connection->recvbuf[11] > '9')
				return false;

		case 11:
			if (connection->recvbuf[10] < '0' || connection->recvbuf[10] > '9')
				return false;

		case 10:
			if (connection->recvbuf[9] != '2')
				return false;

		case 9:
			if (connection->recvbuf[8] != ' ')
				return false;

		case 8:
			if (connection->recvbuf[7] != '0' && connection->recvbuf[7] != '1')
				return false;

		case 7:
			if (connection->recvbuf[6] != '.')
				return false;

		case 6:
			if (connection->recvbuf[5] != '1')
				return false;

		case 5:
			if (connection->recvbuf[4] != '/')
				return false;

		case 4:
			if (connection->recvbuf[3] != 'P')
				return false;

		case 3:
			if (connection->recvbuf[2] != 'T')
				return false;

		case 2:
			if (connection->recvbuf[1] != 'T')
				return false;

		case 1:
			if (connection->recvbuf[0] != 'H')
				return false;
	}

	if (avail < 12)
		return true;

	(void) host->callback(client, host->address_str, port->port, port->type);
	(void) acopm_proxy_host_free(host);
	return true;
}
