/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include <arpa/inet.h>

#include <event2/event.h>

#include "proxy/client.h"
#include "proxy/host.h"
#include "proxy/internal.h"
#include "proxy/port.h"
#include "proxy/protos.h"
#include "utils/bufconn.h"
#include "utils/ignore_errno.h"
#include "utils/log.h"

#include "socks4.h"

static bool acopm_socks4_recv_response(struct acopm_connection *);

bool
acopm_socks4_send_request(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_proxy_scan *const scan = (struct acopm_proxy_scan *) connection->context;

	assert(scan->conn == connection);
	assert(scan->host != NULL);
	assert(scan->port != NULL);
	assert(scan->host->client != NULL);
	assert(scan->port->client != NULL);
	assert(scan->host->client == scan->port->client);

	struct acopm_proxy_host *const host = scan->host;
	struct acopm_proxy_port *const port = scan->port;
	struct acopm_proxy_client *const client = host->client;

	struct acopm_socks4_request request;

	request.version = ACOPM_SOCKS4_VERSION;
	request.command = ACOPM_SOCKS4_CMD_CONNECT;
	request.port = htons(port->port);

	(void) memcpy(&request.addr, &host->address_raw, INET_ADDRLEN);

	if (! acopm_conn_writebuf(connection, &request, sizeof request))
		return false;

	if (! acopm_conn_writebuf(connection, client->username, client->username_length))
		return false;

	return acopm_conn_ev_mod(connection, EV_READ, EV_WRITE, acopm_socks4_recv_response, NULL);
}

static bool
acopm_socks4_recv_response(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_proxy_scan *const scan = (struct acopm_proxy_scan *) connection->context;

	assert(scan->conn == connection);
	assert(scan->host != NULL);
	assert(scan->port != NULL);
	assert(scan->host->client != NULL);
	assert(scan->port->client != NULL);
	assert(scan->host->client == scan->port->client);

	struct acopm_proxy_host *const host = scan->host;
	struct acopm_proxy_port *const port = scan->port;
	struct acopm_proxy_client *const client = host->client;

	struct acopm_socks4_response *const response = (struct acopm_socks4_response *) connection->recvbuf;

	if (! acopm_conn_readbuf(connection, acopm_socks4_recv_response))
		return false;

	if (connection->recvbuf_used < sizeof *response)
		return true;

	if (response->null != 0x00)
		return false;

	switch (response->status)
	{
		case ACOPM_SOCKS4_REQ_GRANTED:
			(void) host->callback(client, host->address_str, port->port, port->type);
			(void) acopm_proxy_host_free(host);
			return true;
		case ACOPM_SOCKS4_REQ_FAILED_IDENT_UNREACH:
			(void) acopm_log_warning("%s: request failed because identd unreachable", __func__);
			(void) acopm_log_warning("%s: check identd is running", __func__);
			break;
		case ACOPM_SOCKS4_REQ_FAILED_IDENT_MISMATCH:
			(void) acopm_log_warning("%s: request failed because identd does not confirm us", __func__);
			(void) acopm_log_warning("%s: check the configuration value of proxyscan::username", __func__);
			break;
		case ACOPM_SOCKS4_REQ_REJECTED:
		default:
			break;
	}

	return false;
}
