/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include <arpa/inet.h>

#include <event2/event.h>

#include "proxy/client.h"
#include "proxy/host.h"
#include "proxy/internal.h"
#include "proxy/port.h"
#include "proxy/protos.h"
#include "utils/bufconn.h"
#include "utils/log.h"

#include "socks5.h"

static bool acopm_socks5_recv_method(struct acopm_connection *);
static bool acopm_socks5_send_request(struct acopm_connection *);
static bool acopm_socks5_recv_response(struct acopm_connection *);

bool
acopm_socks5_send_methods(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);

	static const uint8_t acopm_socks5_authmethods[] = {
		ACOPM_SOCKS5_AUTHMETHOD_NONE
	};

	static const uint8_t acopm_socks5_authmethods_count = 1;

	struct acopm_socks5_methods methods;

	methods.version = ACOPM_SOCKS5_VERSION;
	methods.methods = acopm_socks5_authmethods_count;

	if (! acopm_conn_writebuf(connection, &methods, sizeof methods))
		return false;

	if (! acopm_conn_writebuf(connection, acopm_socks5_authmethods, sizeof acopm_socks5_authmethods))
		return false;

	return acopm_conn_ev_mod(connection, EV_READ, EV_WRITE, acopm_socks5_recv_method, NULL);
}

static bool
acopm_socks5_recv_method(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);

	struct acopm_socks5_method *const method = (struct acopm_socks5_method *) connection->recvbuf;

	if (! acopm_conn_readbuf(connection, acopm_socks5_recv_method))
		return false;

	if (connection->recvbuf_used < sizeof *method)
		return true;

	if (connection->recvbuf_used > sizeof *method)
		return false;

	if (method->version != ACOPM_SOCKS5_VERSION)
		return false;

	if (method->method != ACOPM_SOCKS5_AUTHMETHOD_NONE)
		return false;

	return acopm_conn_ev_mod(connection, EV_WRITE, EV_READ, NULL, acopm_socks5_send_request);
}

static bool
acopm_socks5_send_request(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_proxy_scan *const scan = (struct acopm_proxy_scan *) connection->context;

	assert(scan->conn == connection);
	assert(scan->host != NULL);
	assert(scan->port != NULL);

	struct acopm_proxy_host *const host = scan->host;
	struct acopm_proxy_port *const port = scan->port;

	union acopm_socks5_request request;

	request.rx.version = ACOPM_SOCKS5_VERSION;
	request.rx.command = ACOPM_SOCKS5_CMD_CONNECT;

	if (host->address_family == AF_INET)
	{
		request.r4.rx.addrtype = ACOPM_SOCKS5_ADDRTYPE_IPV4;
		request.r4.port = htons(port->port);

		(void) memcpy(&request.r4.address, host->address_raw, INET_ADDRLEN);

		if (! acopm_conn_writebuf(connection, &request, sizeof request.r4))
			return false;
	}
	else if (host->address_family == AF_INET6)
	{
		request.r6.rx.addrtype = ACOPM_SOCKS5_ADDRTYPE_IPV6;
		request.r6.port = htons(port->port);

		(void) memcpy(&request.r6.address, host->address_raw, INET6_ADDRLEN);

		if (! acopm_conn_writebuf(connection, &request, sizeof request.r6))
			return false;
	}
	else
	{
		(void) acopm_log_critical("%s: unknown address family %d (BUG!)", __func__, (int) host->address_family);
		return false;
	}

	return acopm_conn_ev_mod(connection, EV_READ, EV_WRITE, acopm_socks5_recv_response, NULL);
}

static bool
acopm_socks5_recv_response(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->context != NULL);

	struct acopm_proxy_scan *const scan = (struct acopm_proxy_scan *) connection->context;

	assert(scan->conn == connection);
	assert(scan->host != NULL);
	assert(scan->port != NULL);
	assert(scan->host->client != NULL);
	assert(scan->port->client != NULL);
	assert(scan->host->client == scan->port->client);

	struct acopm_proxy_host *const host = scan->host;
	struct acopm_proxy_port *const port = scan->port;
	struct acopm_proxy_client *const client = host->client;

	struct acopm_socks5_response_vx *const response = (struct acopm_socks5_response_vx *) connection->recvbuf;

	if (! acopm_conn_readbuf(connection, acopm_socks5_recv_response))
		return false;

	if (connection->recvbuf_used < sizeof *response)
		return true;

	if (response->version != ACOPM_SOCKS5_VERSION)
		return false;

	if (response->status != ACOPM_SOCKS5_REQ_GRANTED)
		return false;

	if (host->address_family == AF_INET)
	{
		if (response->addrtype != ACOPM_SOCKS5_ADDRTYPE_IPV4)
			return false;

		struct acopm_socks5_response_v4 *const resp = (struct acopm_socks5_response_v4 *) response;

		if (connection->recvbuf_used < sizeof *resp)
			return true;

		if (memcmp(&resp->address, host->address_raw, INET_ADDRLEN) != 0)
			return false;

		if (resp->port != htons(port->port))
			return false;
	}
	else if (host->address_family == AF_INET6)
	{
		if (response->addrtype != ACOPM_SOCKS5_ADDRTYPE_IPV6)
			return false;

		struct acopm_socks5_response_v6 *const resp = (struct acopm_socks5_response_v6 *) response;

		if (connection->recvbuf_used < sizeof *resp)
			return true;

		if (memcmp(&resp->address, host->address_raw, INET6_ADDRLEN) != 0)
			return false;

		if (resp->port != htons(port->port))
			return false;
	}
	else
	{
		(void) acopm_log_critical("%s: unknown address family %d (BUG!)", __func__, (int) host->address_family);
		return false;
	}

	(void) host->callback(client, host->address_str, port->port, port->type);
	(void) acopm_proxy_host_free(host);
	return true;
}
