/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <string.h> /* memset() */

#include <arpa/inet.h> /* inet_pton() */
#include <netinet/in.h> /* INET6_ADDRSTRLEN */
#include <sys/socket.h> /* struct sockaddr_storage, sa_family_t, AF_* */

#include "utils/addr.h" /* decls for own functions */
#include "utils/log.h" /* acopm_log_error() */

bool
acopm_addrisvalid(const char *const restrict address, sa_family_t *const restrict family)
{
	unsigned char tmp[64];
	sa_family_t fam = AF_UNSPEC;

	if (inet_pton((fam = AF_INET), address, tmp) == 1)
		goto valid;

	if (inet_pton((fam = AF_INET6), address, tmp) == 1)
		goto valid;

	return false;

valid:
	if (family)
		*family = fam;

	return true;
}

bool
acopm_get_sockaddr(const int fd, char buf[static ACOPM_ADDRSTRLEN], const bool remote)
{
	struct sockaddr_storage ss;

	struct sockaddr     *const sax = (struct sockaddr *)     &ss;
	struct sockaddr_in  *const sa4 = (struct sockaddr_in *)  &ss;
	struct sockaddr_in6 *const sa6 = (struct sockaddr_in6 *) &ss;

	(void) memset(&ss, 0x00, sizeof ss);
	socklen_t sax_len = (socklen_t) sizeof ss;

	char addr[INET6_ADDRSTRLEN];

	if (! remote)
	{
		if (getsockname(fd, sax, &sax_len) != 0)
		{
			(void) acopm_log_error("%s: getsockname: %s", __func__, strerror(errno));
			return false;
		}
	}
	else
	{
		if (getpeername(fd, sax, &sax_len) != 0)
		{
			(void) acopm_log_error("%s: getpeername: %s", __func__, strerror(errno));
			return false;
		}
	}

	assert((sax_len == sizeof *sa4) || (sax_len == sizeof *sa6));

	switch (sax->sa_family)
	{
		case AF_INET:

			if (! inet_ntop(sax->sa_family, &sa4->sin_addr, addr, sizeof addr))
			{
				(void) acopm_log_error("%s: inet_ntop: %s", __func__, strerror(errno));
				return false;
			}

			(void) snprintf(buf, ACOPM_ADDRSTRLEN, "%s:%hu", addr, ntohs(sa4->sin_port));
			return true;

		case AF_INET6:

			if (! inet_ntop(sax->sa_family, &sa6->sin6_addr, addr, sizeof addr))
			{
				(void) acopm_log_error("%s: inet_ntop: %s", __func__, strerror(errno));
				return false;
			}

			(void) snprintf(buf, ACOPM_ADDRSTRLEN, "[%s]:%hu", addr, ntohs(sa6->sin6_port));
			return true;
	}

	(void) acopm_log_error("%s: unknown socket address family %d", __func__, sax->sa_family);
	return false;
}
