/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <string.h> /* memset() */
#include <time.h> /* time_t */
#include <unistd.h> /* close() */

#include <arpa/inet.h> /* socklen_t */

#include <event2/event.h> /* struct event_base, event_*() */

#ifdef BUILD_WITH_MBEDTLS
#  include <mbedtls/md.h> /* mbedtls_md_*, MBEDTLS_MD_*, mbedtls_md*() */
#  include <mbedtls/pk.h> /* mbedtls_pk_context, mbedtls_pk_*() */
#  include <mbedtls/ssl.h> /* mbedtls_ssl_*, mbedtls_ssl_*(), MBEDTLS_ERR_SSL_* */
#  include <mbedtls/version.h> /* MBEDTLS_VERSION_NUMBER */
#  include <mbedtls/x509_crt.h> /* mbedtls_x509_crt, mbedtls_x509_*() */
#  define ACOPM_PUBKEY_LEN_MAX 524288U
#endif

#include "3rdparty/utlist.h" /* LL_FOREACH(), LL_FOREACH_SAFE(), LL_DELETE(), LL_PREPEND() */
#include "utils/addr.h" /* acopm_get_sockaddr() */
#include "utils/base64.h" /* acopm_base64encode() */
#include "utils/bufconn.h" /* decls for own functions */
#include "utils/ignore_errno.h" /* acopm_ignore_errno() */
#include "utils/log.h" /* acopm_log_*() */
#include "utils/misc.h" /* acopm_mbed_strerror(), acopm_rng_tls() */

static inline const char * acopm_attr_inline
acopm_conn_name(const struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);

	const char *const connection_name = (*connection->name) ? connection->name : "<unknown>";

	static char result[512];
	(void) snprintf(result, sizeof result, "#%d[%s]", connection->socket_fd, connection_name);
	return result;
}

static inline int acopm_attr_inline
acopm_conn_net_xfer(struct acopm_connection *const restrict connection, const size_t len,
                    void *const restrict recvbuf, const void *const restrict sendbuf)
{
	assert((recvbuf != NULL && sendbuf == NULL) || (recvbuf == NULL && sendbuf != NULL));

	ssize_t ret = 0;

	errno = 0;

	if (recvbuf)
		ret = recv(connection->socket_fd, recvbuf, len, MSG_DONTWAIT);

	if (sendbuf)
		ret = send(connection->socket_fd, sendbuf, len, MSG_DONTWAIT);

	int net_errno = errno;

	if (! net_errno && ret < 0)
		net_errno = EIO;
	if (! net_errno && ret < 1)
		net_errno = ECONNRESET;

	connection->net_errno = net_errno;

#ifdef BUILD_WITH_MBEDTLS
	if (connection->use_tls && ret < 0 && acopm_ignore_errno(net_errno))
		return MBEDTLS_ERR_SSL_WANT_READ;
#endif

	assert(ret <= INT_MAX);
	return (int) ret;
}

static int
acopm_conn_net_recv_func(void *const restrict ctx, void *const restrict buf, const size_t len)
{
	assert(ctx != NULL);
	assert(buf != NULL);
	assert(len > 0);

	return acopm_conn_net_xfer((struct acopm_connection *) ctx, len, buf, NULL);
}

static int
acopm_conn_net_send_func(void *const restrict ctx, const void *const restrict buf, const size_t len)
{
	assert(ctx != NULL);
	assert(buf != NULL);
	assert(len > 0);

	return acopm_conn_net_xfer((struct acopm_connection *) ctx, len, NULL, buf);
}

#ifdef BUILD_WITH_MBEDTLS

static inline bool acopm_attr_inline
acopm_conn_certfp_spki(const enum acopm_conn_certfp_method method, bool *const restrict value)
{
	assert(method != ACOPM_CONN_CERTFP_METHOD_NONE);

	switch (method)
	{
		case ACOPM_CONN_CERTFP_METHOD_NONE:
			break;

		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA1_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA1_HEX:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_256_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_256_HEX:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_512_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_512_HEX:
			*value = true;
			return true;

		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA1_B64:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA1_HEX:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_256_B64:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_256_HEX:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_512_B64:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_512_HEX:
			*value = false;
			return true;
	}

	(void) acopm_log_critical("%s: unknown certificate fingerprint algorithm (BUG)", __func__);
	return false;
}

static inline bool acopm_attr_inline
acopm_conn_certfp_mdtype(const enum acopm_conn_certfp_method method, mbedtls_md_type_t *const restrict value)
{
	assert(method != ACOPM_CONN_CERTFP_METHOD_NONE);

	switch (method)
	{
		case ACOPM_CONN_CERTFP_METHOD_NONE:
			break;

		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA1_B64:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA1_HEX:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA1_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA1_HEX:
			*value = MBEDTLS_MD_SHA1;
			return true;

		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_256_B64:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_256_HEX:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_256_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_256_HEX:
			*value = MBEDTLS_MD_SHA256;
			return true;

		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_512_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_512_B64:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_512_HEX:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_512_HEX:
			*value = MBEDTLS_MD_SHA512;
			return true;
	}

	(void) acopm_log_critical("%s: unknown certificate fingerprint algorithm (BUG)", __func__);
	return false;
}

static inline bool acopm_attr_inline
acopm_conn_certfp_base64(const enum acopm_conn_certfp_method method, bool *const restrict value)
{
	assert(method != ACOPM_CONN_CERTFP_METHOD_NONE);

	switch (method)
	{
		case ACOPM_CONN_CERTFP_METHOD_NONE:
			break;

		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA1_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA1_B64:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_256_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_256_B64:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_512_B64:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_512_B64:
			*value = true;
			return true;

		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA1_HEX:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA1_HEX:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_256_HEX:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_256_HEX:
		case ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_512_HEX:
		case ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_512_HEX:
			*value = false;
			return true;
	}

	(void) acopm_log_critical("%s: unknown certificate fingerprint algorithm (BUG)", __func__);
	return false;
}

#if (MBEDTLS_VERSION_NUMBER < 0xFFFFFFFFUL)
// https://github.com/ARMmbed/mbedtls/issues/396
// TODO XXX when issue resolved, update version number

static const char *
acopm_conn_compute_certfp(const struct acopm_connection *const restrict connection,
                          mbedtls_x509_crt *const restrict certificate)

#else /* MBEDTLS_VERSION_NUMBER */

static const char *
acopm_conn_compute_certfp(const struct acopm_connection *const restrict connection,
                          const mbedtls_x509_crt *const restrict certificate)

#endif /* MBEDTLS_VERSION_NUMBER */
{
	assert(connection != NULL);
	assert(connection->tls != NULL);
	assert(connection->tls->certfp_method != ACOPM_CONN_CERTFP_METHOD_NONE);
	assert(certificate != NULL);

	bool spki = false, base64 = false;
	mbedtls_md_type_t md_type = MBEDTLS_MD_NONE;

	if (! acopm_conn_certfp_spki(connection->tls->certfp_method, &spki))
		return NULL;
	if (! acopm_conn_certfp_mdtype(connection->tls->certfp_method, &md_type))
		return NULL;
	if (! acopm_conn_certfp_base64(connection->tls->certfp_method, &base64))
		return NULL;

	const mbedtls_md_info_t *const md_info = mbedtls_md_info_from_type(md_type);
	if (! md_info)
	{
		(void) acopm_log_critical("%s(%s): md_info_from_type: no algorithm (BUG)",
		                          __func__, acopm_conn_name(connection));
		return NULL;
	}

	const size_t md_length = (size_t) mbedtls_md_get_size(md_info);

	int ret = 0;
	unsigned char publickey_der[ACOPM_PUBKEY_LEN_MAX];

	if (spki && (ret = mbedtls_pk_write_pubkey_der(&certificate->pk, publickey_der, sizeof publickey_der)) < 0)
	{
		(void) acopm_log_error("%s(%s): pk_write_pubkey_der: %s",
		                       __func__, acopm_conn_name(connection), acopm_mbed_strerror(ret));
		return NULL;
	}

	unsigned char fingerprint_der[ACOPM_CONN_CERTFP_LEN_MAX];
	void *const data = ((spki) ? (publickey_der + (sizeof publickey_der - (size_t) ret)) : (certificate->raw.p));
	const size_t len = ((spki) ? ((size_t) ret) : (certificate->raw.len));

	if ((ret = mbedtls_md(md_info, data, len, fingerprint_der)) != 0)
	{
		(void) acopm_log_error("%s(%s): mbedtls_md: %s",
		                       __func__, acopm_conn_name(connection), acopm_mbed_strerror(ret));
		return NULL;
	}

	static char fingerprint_str[(ACOPM_CONN_CERTFP_LEN_MAX * 2) + 4];

	if (base64)
		(void) acopm_base64encode(fingerprint_der, md_length, fingerprint_str);
	else
		for (size_t x = 0; x < md_length; x++)
			(void) sprintf(fingerprint_str + (x * 2), "%02x", fingerprint_der[x]);

	return fingerprint_str;
}

static bool
acopm_conn_verify_certfp(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->tls != NULL);

	if (connection->tls->cert_verification_success)
		return true;

	if (! connection->tls->certfp_values)
		return false;

	const mbedtls_x509_crt *const certificate = mbedtls_ssl_get_peer_cert(&connection->tls->tls_ctx);

	if (! certificate)
	{
		(void) acopm_log_error("%s(%s): ssl_get_peer_cert: no certificate",
		                       __func__, acopm_conn_name(connection));
		return false;
	}

#if (MBEDTLS_VERSION_NUMBER < 0xFFFFFFFFUL)
	// https://github.com/ARMmbed/mbedtls/issues/396
	// TODO XXX when issue resolved, update version number

	int ret;
	mbedtls_x509_crt tmp;

	(void) mbedtls_x509_crt_init(&tmp);

	if ((ret = mbedtls_x509_crt_parse_der(&tmp, certificate->raw.p, certificate->raw.len)) != 0)
	{
		(void) acopm_log_error("%s(%s): x509_crt_parse_der: %s",
		                       __func__, acopm_conn_name(connection), acopm_mbed_strerror(ret));
		(void) mbedtls_x509_crt_free(&tmp);
		return false;
	}

	const char *const certfp_value = acopm_conn_compute_certfp(connection, &tmp);

	(void) mbedtls_x509_crt_free(&tmp);
#else /* MBEDTLS_VERSION_NUMBER */
	const char *const certfp_value = acopm_conn_compute_certfp(connection, certificate);
#endif /* MBEDTLS_VERSION_NUMBER */

	if (! certfp_value)
		return false;

	(void) acopm_log_debug("%s(%s): computed server fingerprint %s", __func__, acopm_conn_name(connection),
	                       certfp_value);

	const struct acopm_conn_certfp_value *cur;
	LL_FOREACH(connection->tls->certfp_values, cur)
	{
		(void) acopm_log_debug("%s(%s): testing fingerprint against %s", __func__,
		                       acopm_conn_name(connection), cur->value);

		if (strcmp(cur->value, certfp_value) == 0)
		{
			(void) acopm_log_debug("%s(%s): certificate fingerprint match", __func__,
			                       acopm_conn_name(connection));

			connection->tls->cert_verification_success = true;
			return true;
		}

		(void) acopm_log_debug("%s(%s): certificate fingerprint mismatch", __func__,
		                       acopm_conn_name(connection));
	}

	(void) acopm_log_error("%s(%s): certificate fingerprint mismatch", __func__, acopm_conn_name(connection));
	return false;
}

static bool
acopm_conn_init_tls(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);

	if (! connection->tls)
		return false;

	int ret;

	mbedtls_x509_crt *const tls_crt = &connection->tls->tls_crt;
	mbedtls_pk_context *const tls_key = &connection->tls->tls_key;
	mbedtls_ssl_config *const tls_cfg = &connection->tls->tls_cfg;
	mbedtls_ssl_context *const tls_ctx = &connection->tls->tls_ctx;

	const char *const cc_file = connection->tls->cc_file;
	const char *const ck_file = connection->tls->ck_file;

#define ACOPM_MBEDTLS_FLAGS     MBEDTLS_SSL_IS_CLIENT, MBEDTLS_SSL_TRANSPORT_STREAM, MBEDTLS_SSL_PRESET_DEFAULT

	if ((ret = mbedtls_ssl_config_defaults(tls_cfg, ACOPM_MBEDTLS_FLAGS)) != 0)
	{
		(void) acopm_log_error("%s(%s): ssl_config_defaults: %s",
		                       __func__, connection->name, acopm_mbed_strerror(ret));
		return false;
	}

	(void) mbedtls_ssl_conf_authmode(tls_cfg, MBEDTLS_SSL_VERIFY_NONE);
	(void) mbedtls_ssl_conf_rng(tls_cfg, acopm_rng_tls, NULL);

#ifdef MBEDTLS_SSL_SESSION_TICKETS
	(void) mbedtls_ssl_conf_session_tickets(tls_cfg, MBEDTLS_SSL_SESSION_TICKETS_DISABLED);
#endif

	if (cc_file && *cc_file && ck_file && *ck_file)
	{
		if ((ret = mbedtls_x509_crt_parse_file(tls_crt, cc_file)) != 0)
		{
			(void) acopm_log_error("%s(%s): parse_crtfile: %s",
			                       __func__, connection->name, acopm_mbed_strerror(ret));
			return false;
		}
		if ((ret = mbedtls_pk_parse_keyfile(tls_key, ck_file, NULL)) != 0)
		{
			(void) acopm_log_error("%s(%s): parse_keyfile: %s",
			                       __func__, connection->name, acopm_mbed_strerror(ret));
			return false;
		}
		if ((ret = mbedtls_pk_check_pair(&tls_crt->pk, tls_key)) != 0)
		{
			(void) acopm_log_error("%s(%s): pk_check_pair: %s",
			                       __func__, connection->name, acopm_mbed_strerror(ret));
			return false;
		}
		if ((ret = mbedtls_ssl_conf_own_cert(tls_cfg, tls_crt, tls_key)) != 0)
		{
			(void) acopm_log_error("%s(%s): conf_own_cert: %s",
			                       __func__, connection->name, acopm_mbed_strerror(ret));
			return false;
		}
	}

	if ((ret = mbedtls_ssl_setup(tls_ctx, tls_cfg)) != 0)
	{
		(void) acopm_log_error("%s(%s): ssl_setup: %s", __func__, connection->name, acopm_mbed_strerror(ret));
		return false;
	}

	mbedtls_ssl_send_t *const send_func = (mbedtls_ssl_send_t *) acopm_conn_net_send_func;
	mbedtls_ssl_recv_t *const recv_func = (mbedtls_ssl_recv_t *) acopm_conn_net_recv_func;

	(void) mbedtls_ssl_set_bio(tls_ctx, connection, send_func, recv_func, NULL);

	return true;
}

static bool
acopm_conn_tls_handshake_cb(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->tls != NULL);

	(void) acopm_log_debug("%s(%s): handshaking ...", __func__, acopm_conn_name(connection));

	const int ret = mbedtls_ssl_handshake(&connection->tls->tls_ctx);

	if (ret == MBEDTLS_ERR_SSL_WANT_READ)
		return acopm_conn_ev_mod(connection, EV_READ, EV_WRITE, acopm_conn_tls_handshake_cb, NULL);

	if (ret == MBEDTLS_ERR_SSL_WANT_WRITE)
		return acopm_conn_ev_mod(connection, EV_WRITE, EV_READ, NULL, acopm_conn_tls_handshake_cb);

	if (ret != 0)
	{
		if (connection->net_errno)
			(void) acopm_log_error("%s(%s): TLS handshake failed: %s", __func__,
			                       acopm_conn_name(connection), strerror(connection->net_errno));
		else
			(void) acopm_log_error("%s(%s): TLS handshake failed: %s", __func__,
			                       acopm_conn_name(connection), acopm_mbed_strerror(ret));

		return false;
	}

	if (connection->tls->certfp_method != ACOPM_CONN_CERTFP_METHOD_NONE)
		if (! acopm_conn_verify_certfp(connection))
			return false;

	(void) acopm_log_debug("%s(%s): TLS handshake complete", __func__, acopm_conn_name(connection));

	const short ev_on = connection->ev_es;
	const short ev_off = ev_on ^ (EV_READ | EV_WRITE);

	return acopm_conn_ev_mod(connection, ev_on, ev_off, connection->ev_rfs, connection->ev_sfs);
}

static int
acopm_conn_tls_recv_func(void *const restrict ctx, void *const restrict buf, const size_t len)
{
	struct acopm_connection *const connection = (struct acopm_connection *) ctx;

	assert(connection != NULL);
	assert(connection->tls != NULL);
	assert(buf != NULL);
	assert(len > 0);

	return mbedtls_ssl_read(&connection->tls->tls_ctx, buf, len);
}

static int
acopm_conn_tls_send_func(void *const restrict ctx, const void *const restrict buf, const size_t len)
{
	struct acopm_connection *const connection = (struct acopm_connection *) ctx;

	assert(connection != NULL);
	assert(connection->tls != NULL);
	assert(buf != NULL);
	assert(len > 0);

	return mbedtls_ssl_write(&connection->tls->tls_ctx, buf, len);
}

static inline void acopm_attr_inline
acopm_conn_ev_stash(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);

	if (connection->stashed)
		return;

	connection->ev_rfs = connection->ev_rf;
	connection->ev_sfs = connection->ev_sf;
	connection->ev_es = connection->ev_e;
	connection->stashed = true;
}

static inline bool acopm_attr_inline
acopm_conn_ev_unstash(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);

	if (! connection->stashed)
		return true;

	connection->stashed = false;

	const short ev_on = connection->ev_es;
	const short ev_off = ev_on ^ (EV_READ | EV_WRITE);

	return acopm_conn_ev_mod(connection, ev_on, ev_off, connection->ev_rfs, connection->ev_sfs);
}

#endif /* BUILD_WITH_MBEDTLS */

static void
acopm_conn_free_internal(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);

	if (connection->connected)
		(void) acopm_log_debug("%s(%s): Disconnecting", __func__, acopm_conn_name(connection));

	if (connection->connected && connection->ev_break && connection->ev_base)
	{
		(void) acopm_log_notice("Breaking event loop");
		(void) event_base_loopbreak(connection->ev_base);
		connection->ev_break = false;
	}

#ifdef BUILD_WITH_MBEDTLS
	if (connection->tls)
	{
		(void) mbedtls_ssl_free(&connection->tls->tls_ctx);
		(void) mbedtls_ssl_config_free(&connection->tls->tls_cfg);
		(void) mbedtls_pk_free(&connection->tls->tls_key);
		(void) mbedtls_x509_crt_free(&connection->tls->tls_crt);

		(void) free(connection->tls);
		connection->tls = NULL;
	}
#endif

	if (connection->socket_fd > -1)
	{
		(void) close(connection->socket_fd);
		connection->socket_fd = -1;
	}
	if (connection->ev_r)
	{
		(void) event_free(connection->ev_r);
		connection->ev_r = NULL;
	}
	if (connection->ev_s)
	{
		(void) event_free(connection->ev_s);
		connection->ev_s = NULL;
	}
	if (connection->ev_x)
	{
		(void) event_free(connection->ev_x);
		connection->ev_x = NULL;
	}
	if (connection->recvbuf)
	{
		(void) free(connection->recvbuf);
		connection->recvbuf = NULL;
	}
	if (connection->sendbuf)
	{
		(void) free(connection->sendbuf);
		connection->sendbuf = NULL;
	}

	connection->connected = false;
}

static bool
acopm_conn_sendbuf_resize(struct acopm_connection *const restrict connection, const size_t sendbuf_size_new)
{
	assert(connection != NULL);
	assert(sendbuf_size_new >= connection->sendbuf_used);

	void *const sendbuf_new = realloc(connection->sendbuf, sendbuf_size_new);

	if (! sendbuf_new)
	{
		(void) acopm_log_error("%s(%s): realloc: %s", __func__, acopm_conn_name(connection), strerror(errno));
		return false;
	}

	connection->sendbuf = sendbuf_new;
	connection->sendbuf_size = sendbuf_size_new;

	return true;
}

static bool
acopm_conn_xmitbuf_out(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);
	assert(connection->sendbuf != NULL);
	assert(connection->sendbuf_used > 0);
	assert(connection->net_send_func != NULL);

	(void) acopm_log_debug("%s(%s): emptying send buffer", __func__, acopm_conn_name(connection));

	const unsigned char *const buf = connection->sendbuf;
	const size_t len = connection->sendbuf_used;
	const int ret = connection->net_send_func(connection, buf, len);

	if (ret <= 0)
	{
#ifdef BUILD_WITH_MBEDTLS
		if (connection->use_tls && ret == MBEDTLS_ERR_SSL_WANT_READ)
		{
			(void) acopm_conn_ev_stash(connection);
			return acopm_conn_ev_mod(connection, EV_READ, EV_WRITE, acopm_conn_xmitbuf_out, NULL);
		}
		if (connection->use_tls && ret == MBEDTLS_ERR_SSL_WANT_WRITE)
		{
			(void) acopm_conn_ev_stash(connection);
			return acopm_conn_ev_mod(connection, EV_WRITE, EV_READ, NULL, acopm_conn_xmitbuf_out);
		}
#endif
		if (acopm_ignore_errno(connection->net_errno))
		{
			(void) acopm_log_debug("%s(%s): send: %s", __func__,
			                       acopm_conn_name(connection), strerror(connection->net_errno));

			return true;
		}
		else if (connection->net_errno)
			(void) acopm_log_error("%s(%s): send: %s", __func__,
			                       acopm_conn_name(connection), strerror(connection->net_errno));
#ifdef BUILD_WITH_MBEDTLS
		else if (connection->use_tls)
			(void) acopm_log_error("%s(%s): send: %s", __func__,
			                       acopm_conn_name(connection), acopm_mbed_strerror(ret));
#endif
		return false;
	}

#ifdef BUILD_WITH_MBEDTLS
	if (! acopm_conn_ev_unstash(connection))
		return false;
#endif

	const size_t data_sent = (size_t) ret;

	(void) acopm_log_debug("%s(%s): sent %zu/%zu bytes to server",
	                       __func__, acopm_conn_name(connection), data_sent, connection->sendbuf_used);

	if ((connection->sendbuf_used -= data_sent))
	{
		unsigned char *const sendbuf_data = connection->sendbuf + data_sent;

		(void) acopm_log_debug("%s(%s): moving %zu bytes from %zu to start of send buffer",
		                       __func__, acopm_conn_name(connection), connection->sendbuf_used, data_sent);

		(void) memmove(connection->sendbuf, sendbuf_data, connection->sendbuf_used);
	}

	(void) acopm_log_debug("%s(%s): now have %zu/%zu bytes in send buffer", __func__,
	                       acopm_conn_name(connection), connection->sendbuf_used, connection->sendbuf_size);

	if (! connection->sendbuf_used)
	{
		if (connection->sendbuf_size > connection->sendbuf_orig)
		{
			(void) acopm_log_debug("%s(%s): collapsing send buffer (%zu -> %zu bytes)", __func__,
			                       acopm_conn_name(connection), connection->sendbuf_size,
			                       connection->sendbuf_orig);

			(void) acopm_conn_sendbuf_resize(connection, connection->sendbuf_orig);
		}

		(void) acopm_log_debug("%s(%s): work done", __func__, acopm_conn_name(connection));

		return acopm_conn_ev_mod(connection, EV_NONE, EV_WRITE, connection->ev_rf, connection->ev_sf);
	}

	return true;
}

static void
acopm_conn_recv_cb(const evutil_socket_t acopm_attr_unused socket_fd, const short events, void *const restrict ctx)
{
	assert(events == EV_READ);
	assert(ctx != NULL);

	(void) events;

	struct acopm_connection *const connection = (struct acopm_connection *) ctx;

	assert((connection->ev_e & events) == events);

	bool result = false;

	if (connection->ev_rf)
		result = connection->ev_rf(connection);
	else
		result = acopm_conn_ev_mod(connection, EV_NONE, EV_READ, connection->ev_rf, connection->ev_sf);

	if (! result)
	{
		(void) acopm_conn_free_internal(connection);

		if (connection->ev_ef)
			(void) connection->ev_ef(connection, ACOPM_CONN_RECVCB_ERROR);
	}
}

static void
acopm_conn_send_cb(const evutil_socket_t acopm_attr_unused socket_fd, const short events, void *const restrict ctx)
{
	assert(events == EV_WRITE);
	assert(ctx != NULL);

	(void) events;

	struct acopm_connection *const connection = (struct acopm_connection *) ctx;

	assert((connection->ev_e & events) == events);

	bool result = false;

	if (connection->sendbuf_used)
		result = acopm_conn_xmitbuf_out(connection);
	else if (connection->ev_sf)
		result = connection->ev_sf(connection);
	else
		result = acopm_conn_ev_mod(connection, EV_NONE, EV_WRITE, connection->ev_rf, connection->ev_sf);

	if (! result)
	{
		(void) acopm_conn_free_internal(connection);

		if (connection->ev_ef)
			(void) connection->ev_ef(connection, ACOPM_CONN_SENDCB_ERROR);
	}
}

static void
acopm_conn_time_cb(const int acopm_attr_unused unused, const short events, void *const restrict ctx)
{
	assert(events == EV_TIMEOUT);
	assert(ctx != NULL);

	(void) events;

	struct acopm_connection *const connection = (struct acopm_connection *) ctx;

	if (connection->ev_ef)
		(void) connection->ev_ef(connection, ACOPM_CONN_DATA_TIMEOUT);

	(void) acopm_conn_free_internal(connection);
}

static bool
acopm_conn_setup(struct acopm_connection *const restrict connection)
{
	struct event_base *const ev_base = connection->ev_base;
	const short ev_rflag = EV_PERSIST | EV_READ;
	const short ev_wflag = EV_PERSIST | EV_WRITE;

	connection->sendbuf_orig = connection->sendbuf_size;
	connection->connected = true;

	if (! (connection->recvbuf = calloc(1, connection->recvbuf_size)))
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): calloc: %s", __func__, acopm_conn_name(connection), strerror(errno));
		return false;
	}
	if (! (connection->sendbuf = calloc(1, connection->sendbuf_size)))
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): calloc: %s", __func__, acopm_conn_name(connection), strerror(errno));
		return false;
	}

	if (! (connection->ev_r = event_new(ev_base, connection->socket_fd, ev_rflag, acopm_conn_recv_cb, connection)))
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): event_new: %s", __func__, acopm_conn_name(connection), strerror(errno));
		return false;
	}
	if (! (connection->ev_s = event_new(ev_base, connection->socket_fd, ev_wflag, acopm_conn_send_cb, connection)))
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): event_new: %s", __func__, acopm_conn_name(connection), strerror(errno));
		return false;
	}

	if (connection->data_timeout)
	{
		struct timeval timeout;
		(void) memset(&timeout, 0x00, sizeof timeout);
		timeout.tv_sec = connection->data_timeout;

		if (! (connection->ev_x = event_new(ev_base, -1, 0, acopm_conn_time_cb, connection)))
		{
			connection->net_errno = errno;

			(void) acopm_log_error("%s(%s): event_new: %s", __func__, acopm_conn_name(connection),
			                       strerror(errno));
			return false;
		}
		if (event_add(connection->ev_x, &timeout) == -1)
		{
			connection->net_errno = errno;

			(void) acopm_log_error("%s(%s): event_add: %s", __func__, acopm_conn_name(connection),
			                       strerror(errno));
			return false;
		}
	}

#ifdef BUILD_WITH_MBEDTLS
	if (connection->use_tls)
	{
		if (! acopm_conn_init_tls(connection))
			return false;

		connection->net_recv_func = acopm_conn_tls_recv_func;
		connection->net_send_func = acopm_conn_tls_send_func;

		const short ev_on = EV_WRITE;

		return acopm_conn_ev_mod(connection, ev_on, EV_NONE, NULL, acopm_conn_tls_handshake_cb);
	}
#endif

	connection->net_recv_func = acopm_conn_net_recv_func;
	connection->net_send_func = acopm_conn_net_send_func;

	const short ev_on = connection->ev_es;

	return acopm_conn_ev_mod(connection, ev_on, EV_NONE, connection->ev_rfs, connection->ev_sfs);
}

static void
acopm_conn_done_cb(const evutil_socket_t socket_fd, const short events, void *const restrict ctx)
{
	struct acopm_connection *const connection = (struct acopm_connection *) ctx;

	assert(connection != NULL);
	assert(connection->ev_s != NULL);

	(void) event_free(connection->ev_s);
	connection->ev_s = NULL;

	if (events == EV_TIMEOUT)
	{
		connection->net_errno = ETIMEDOUT;
		(void) acopm_log_info("%s(%s): connect: %s", __func__, acopm_conn_name(connection), strerror(ETIMEDOUT));
		(void) acopm_conn_free_internal(connection);

		if (connection->ev_ef)
			(void) connection->ev_ef(connection, ACOPM_CONN_CONNECT_TIMEOUT);

		return;
	}

	int sockerr = 0;
	socklen_t sockerr_sz = (socklen_t) sizeof sockerr;

	if (getsockopt(socket_fd, SOL_SOCKET, SO_ERROR, &sockerr, &sockerr_sz) == -1)
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): getsockopt: %s", __func__, acopm_conn_name(connection), strerror(errno));
		(void) acopm_conn_free_internal(connection);

		if (connection->ev_ef)
			(void) connection->ev_ef(connection, ACOPM_CONN_INTERNAL_ERROR);

		return;
	}
	if (sockerr != 0)
	{
		connection->net_errno = sockerr;
		(void) acopm_log_error("%s(%s): connect: %s", __func__, acopm_conn_name(connection), strerror(sockerr));
		(void) acopm_conn_free_internal(connection);

		if (connection->ev_ef)
			(void) connection->ev_ef(connection, ACOPM_CONN_CONNECT_ERROR);

		return;
	}
	if (! acopm_conn_setup(connection))
	{
		(void) acopm_conn_free_internal(connection);

		if (connection->ev_ef)
			(void) connection->ev_ef(connection, ACOPM_CONN_INTERNAL_ERROR);
	}
}

static bool
acopm_conn_setup_nonblock(struct acopm_connection *const restrict connection)
{
	struct timeval timeout;
	(void) memset(&timeout, 0x00, sizeof timeout);
	timeout.tv_sec = connection->connect_timeout;

	struct event_base *const ev_base = connection->ev_base;
	const short ev_wflag = EV_WRITE;

	if (! (connection->ev_s = event_new(ev_base, connection->socket_fd, ev_wflag, acopm_conn_done_cb, connection)))
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): event_new: %s", __func__, acopm_conn_name(connection), strerror(errno));
		return false;
	}
	if (event_add(connection->ev_s, &timeout) == -1)
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): event_add: %s", __func__, acopm_conn_name(connection), strerror(errno));
		return false;
	}

	return true;
}

struct acopm_connection *
acopm_conn_new(const char *const restrict connection_name, void *const restrict context)
{
	struct acopm_connection *const connection = calloc(1, sizeof *connection);

	if (! connection)
	{
		(void) acopm_log_error("%s(%s): calloc: %s", __func__, connection_name, strerror(errno));
		return NULL;
	}

	connection->context = context;
	connection->socket_fd = -1;

	connection->recvbuf_size = ACOPM_CONN_RECVBUF_LEN_DEFAULT;
	connection->sendbuf_size = ACOPM_CONN_SENDBUF_LEN_DEFAULT;
	connection->sendbuf_orig = ACOPM_CONN_SENDBUF_LEN_DEFAULT;

	(void) acopm_strset_buf(connection->name, connection_name);

	return connection;
}

void
acopm_conn_set_ev_break(struct acopm_connection *const restrict connection, const bool ev_break)
{
	assert(connection != NULL);

	connection->ev_break = ev_break;
}

void
acopm_conn_set_nonblock(struct acopm_connection *const restrict connection, const bool nonblock,
                        const time_t connect_timeout)
{
	assert(connection != NULL);

	connection->nonblock = nonblock;
	connection->connect_timeout = connect_timeout;
}

void
acopm_conn_set_data_cb(struct acopm_connection *const restrict connection, const short ev_es,
                       const acopm_conn_data_func ev_rfs, const acopm_conn_data_func ev_sfs)
{
	assert(connection != NULL);

	connection->ev_es = ev_es;
	connection->ev_rfs = ev_rfs;
	connection->ev_sfs = ev_sfs;
}

void
acopm_conn_set_error_cb(struct acopm_connection *const restrict connection, const acopm_conn_error_func ev_ef)
{
	assert(connection != NULL);

	connection->ev_ef = ev_ef;
}

void
acopm_conn_set_data_timeout(struct acopm_connection *const restrict connection, const time_t data_timeout)
{
	assert(connection != NULL);

	connection->data_timeout = data_timeout;
}

void
acopm_conn_set_bufsz(struct acopm_connection *const restrict connection, const size_t recvsz, const size_t sendsz)
{
	assert(connection != NULL);

	if (connection->connected)
		return;

	connection->recvbuf_size = recvsz;
	connection->sendbuf_size = sendsz;
}

void
acopm_conn_set_bind4(struct acopm_connection *const restrict connection,
                     const struct sockaddr_in *const restrict addr)
{
	assert(connection != NULL);

	connection->bind4 = addr;
}

void
acopm_conn_set_bind6(struct acopm_connection *const restrict connection,
                     const struct sockaddr_in6 *const restrict addr)
{
	assert(connection != NULL);

	connection->bind6 = addr;
}

#ifdef BUILD_WITH_MBEDTLS

bool
acopm_conn_set_tls(struct acopm_connection *const restrict connection, const bool use_tls,
                   const enum acopm_conn_certfp_method certfp_method,
                   const struct acopm_conn_certfp_value *const restrict certfp_values)
{
	assert(connection != NULL);

	assert(
		(certfp_method == ACOPM_CONN_CERTFP_METHOD_NONE) ||
		(certfp_method >= ACOPM_CONN_CERTFP_METHOD_RANGE_MIN && certfp_method <= ACOPM_CONN_CERTFP_METHOD_RANGE_MAX)
	);

	if (connection->connected)
		return false;

	if (connection->tls)
	{
		(void) mbedtls_ssl_free(&connection->tls->tls_ctx);
		(void) mbedtls_ssl_config_free(&connection->tls->tls_cfg);
		(void) mbedtls_pk_free(&connection->tls->tls_key);
		(void) mbedtls_x509_crt_free(&connection->tls->tls_crt);
	}
	if (! (connection->use_tls = use_tls))
	{
		if (connection->tls)
		{
			(void) free(connection->tls);
			connection->tls = NULL;
		}

		return true;
	}
	if (! connection->tls && ! (connection->tls = calloc(1, sizeof *connection->tls)))
	{
		(void) acopm_log_error("%s(%s): calloc: %s", __func__, connection->name, strerror(errno));
		(void) free(connection);
		return false;
	}

	(void) mbedtls_x509_crt_init(&connection->tls->tls_crt);
	(void) mbedtls_pk_init(&connection->tls->tls_key);
	(void) mbedtls_ssl_config_init(&connection->tls->tls_cfg);
	(void) mbedtls_ssl_init(&connection->tls->tls_ctx);

	connection->tls->certfp_method = certfp_method;
	connection->tls->certfp_values = certfp_values;

	return true;
}

void
acopm_conn_set_tls_cert(struct acopm_connection *const restrict connection,
                        const char *const restrict cc_file, const char *const restrict ck_file)
{
	assert(connection != NULL);

	if (! connection->tls || ! connection->use_tls || connection->connected)
		return;

	if (cc_file && *cc_file && ck_file && *ck_file)
	{
		if (! acopm_strset_buf(connection->tls->cc_file, cc_file))
			(void) acopm_log_warning("%s: certificate file '%s' too long, truncated", __func__, cc_file);

		if (! acopm_strset_buf(connection->tls->ck_file, ck_file))
			(void) acopm_log_warning("%s: private key file '%s' too long, truncated", __func__, ck_file);
	}
	else
	{
		(void) memset(connection->tls->cc_file, 0x00, sizeof connection->tls->cc_file);
		(void) memset(connection->tls->ck_file, 0x00, sizeof connection->tls->ck_file);
	}
}

#endif /* BUILD_WITH_MBEDTLS */

bool
acopm_conn_do_connect(struct acopm_connection *const restrict connection, struct event_base *const restrict ev_base,
                      struct sockaddr *const restrict addr, const socklen_t addrlen)
{
	assert(connection != NULL);
	assert(ev_base != NULL);

	assert(addr != NULL);
	assert((addr->sa_family == AF_INET) || (addr->sa_family == AF_INET6));
	assert((addrlen == sizeof(struct sockaddr_in)) || (addrlen == sizeof(struct sockaddr_in6)));

	connection->ev_base = ev_base;

	const int socktype = SOCK_STREAM | SOCK_CLOEXEC | ((connection->nonblock) ? SOCK_NONBLOCK : 0);
	const int protocol = IPPROTO_TCP;

	if ((connection->socket_fd = socket(addr->sa_family, socktype, protocol)) == -1)
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): socket: %s", __func__, connection->name, strerror(errno));
		(void) acopm_conn_free_internal(connection);
		return false;
	}

	const struct sockaddr *sa_bind = NULL;
	socklen_t sa_bind_len = 0;

	if (addr->sa_family == AF_INET && connection->bind4)
	{
		sa_bind = (const struct sockaddr *) connection->bind4;
		sa_bind_len = (socklen_t) (sizeof *connection->bind4);
	}
	else if (addr->sa_family == AF_INET6 && connection->bind6)
	{
		sa_bind = (const struct sockaddr *) connection->bind6;
		sa_bind_len = (socklen_t) (sizeof *connection->bind6);
	}

	if (sa_bind && sa_bind_len && bind(connection->socket_fd, sa_bind, sa_bind_len) == -1)
	{
		connection->net_errno = errno;
		(void) acopm_log_error("%s(%s): bind: %s", __func__, acopm_conn_name(connection), strerror(errno));
		(void) acopm_conn_free_internal(connection);
		return false;
	}

	if (connect(connection->socket_fd, addr, addrlen) == -1 && (! connection->nonblock || errno != EINPROGRESS))
	{
		connection->net_errno = errno;

		if (connection->nonblock)
			(void) acopm_log_debug("%s(%s): connect: %s", __func__, acopm_conn_name(connection),
			                       strerror(errno));
		else
			(void) acopm_log_error("%s(%s): connect: %s", __func__, acopm_conn_name(connection),
			                       strerror(errno));

		(void) acopm_conn_free_internal(connection);
		return false;
	}

	if (connection->nonblock)
	{
		if (! acopm_conn_setup_nonblock(connection))
		{
			(void) acopm_conn_free_internal(connection);
			return false;
		}
	}
	else
	{
		if (! acopm_conn_setup(connection))
		{
			(void) acopm_conn_free_internal(connection);
			return false;
		}
	}

	return true;
}

bool
acopm_conn_get_sockaddr(struct acopm_connection *const restrict connection,
                        char al[static ACOPM_ADDRSTRLEN], char ar[static ACOPM_ADDRSTRLEN])
{
	assert(connection != NULL);
	assert(connection->socket_fd != -1);
	assert(al != NULL);
	assert(ar != NULL);

	if (! acopm_get_sockaddr(connection->socket_fd, al, false))
		return false;

	if (! acopm_get_sockaddr(connection->socket_fd, ar, true))
		return false;

	return true;
}

bool
acopm_conn_ev_mod(struct acopm_connection *const restrict connection,
                  const short ev_on, const short ev_off,
                  const acopm_conn_data_func ev_rf, const acopm_conn_data_func ev_sf)
{
	assert(connection != NULL);
	assert((ev_on & ev_off) == EV_NONE);
	assert((ev_on == EV_NONE) || (ev_on == EV_READ) || (ev_on == EV_WRITE) || (ev_on == (EV_READ | EV_WRITE)));
	assert((ev_off == EV_NONE) || (ev_off == EV_READ) || (ev_off == EV_WRITE));

	if ((ev_on & EV_READ) && ! (connection->ev_e & EV_READ))
	{
		(void) acopm_log_debug("%s(%s): enabling recv events", __func__, acopm_conn_name(connection));

		if (event_add(connection->ev_r, NULL) == -1)
		{
			connection->net_errno = errno;
			(void) acopm_log_error("%s(%s): event_add: %s",
			                       __func__, acopm_conn_name(connection), strerror(errno));
			return false;
		}

		connection->ev_e |= EV_READ;
	}
	if ((ev_off & EV_READ) && (connection->ev_e & EV_READ))
	{
		(void) acopm_log_debug("%s(%s): disabling recv events", __func__, acopm_conn_name(connection));

		if (event_del(connection->ev_r) == -1)
		{
			connection->net_errno = errno;
			(void) acopm_log_error("%s(%s): event_del: %s",
			                       __func__, acopm_conn_name(connection), strerror(errno));
			return false;
		}

		connection->ev_e &= ((short) ~(EV_READ));
	}
	if (connection->ev_rf != ev_rf)
	{
		if (ev_rf)
			(void) acopm_log_debug("%s(%s): changing recv handler", __func__, acopm_conn_name(connection));
		else
			(void) acopm_log_debug("%s(%s): disabling recv handler", __func__, acopm_conn_name(connection));

		connection->ev_rf = ev_rf;
	}
	if ((ev_on & EV_WRITE) && ! (connection->ev_e & EV_WRITE))
	{
		(void) acopm_log_debug("%s(%s): enabling send events", __func__, acopm_conn_name(connection));

		if (event_add(connection->ev_s, NULL) == -1)
		{
			connection->net_errno = errno;
			(void) acopm_log_error("%s(%s): event_add: %s",
			                       __func__, acopm_conn_name(connection), strerror(errno));
			return false;
		}

		connection->ev_e |= EV_WRITE;
	}
	if ((ev_off & EV_WRITE) && (connection->ev_e & EV_WRITE) && (! connection->sendbuf_used))
	{
		(void) acopm_log_debug("%s(%s): disabling send events", __func__, acopm_conn_name(connection));

		if (event_del(connection->ev_s) == -1)
		{
			connection->net_errno = errno;
			(void) acopm_log_error("%s(%s): event_del: %s",
			                       __func__, acopm_conn_name(connection), strerror(errno));
			return false;
		}

		connection->ev_e &= ((short) ~(EV_WRITE));
	}
	if (connection->ev_sf != ev_sf)
	{
		if (ev_sf)
			(void) acopm_log_debug("%s(%s): changing send handler", __func__, acopm_conn_name(connection));
		else
			(void) acopm_log_debug("%s(%s): disabling send handler", __func__, acopm_conn_name(connection));

		connection->ev_sf = ev_sf;
	}

	return true;
}

bool
acopm_conn_readbuf(struct acopm_connection *const restrict connection, const acopm_conn_data_func callback)
{
	assert(connection != NULL);
	assert(connection->recvbuf != NULL);
	assert(connection->net_recv_func != NULL);

	(void) callback;

	unsigned char *const buf = connection->recvbuf + connection->recvbuf_used;
	const size_t len = connection->recvbuf_size - connection->recvbuf_used;

	if (! len)
	{
		(void) acopm_log_warning("%s(%s): trying to read into a full buffer", __func__,
		                         acopm_conn_name(connection));
		return false;
	}

	const int ret = connection->net_recv_func(connection, buf, len);

	if (ret <= 0)
	{
#ifdef BUILD_WITH_MBEDTLS
		if (connection->use_tls && ret == MBEDTLS_ERR_SSL_WANT_READ)
		{
			(void) acopm_conn_ev_stash(connection);
			return acopm_conn_ev_mod(connection, EV_READ, EV_WRITE, callback, NULL);
		}
		if (connection->use_tls && ret == MBEDTLS_ERR_SSL_WANT_WRITE)
		{
			(void) acopm_conn_ev_stash(connection);
			return acopm_conn_ev_mod(connection, EV_WRITE, EV_READ, NULL, callback);
		}
#endif
		if (acopm_ignore_errno(connection->net_errno))
		{
			(void) acopm_log_debug("%s(%s): recv: %s", __func__, acopm_conn_name(connection),
			                       strerror(connection->net_errno));
			return true;
		}
		else if (connection->net_errno)
			(void) acopm_log_error("%s(%s): recv: %s", __func__, acopm_conn_name(connection),
			                       strerror(connection->net_errno));
#ifdef BUILD_WITH_MBEDTLS
		else if (connection->use_tls)
			(void) acopm_log_error("%s(%s): mbedtls_ssl_read: %s", __func__, acopm_conn_name(connection),
			                       acopm_mbed_strerror(ret));
#endif
		return false;
	}

#ifdef BUILD_WITH_MBEDTLS
	if (! acopm_conn_ev_unstash(connection))
		return false;
#endif

	const size_t retsz = (size_t) ret;

	connection->recvbuf_acqd += retsz;
	connection->recvbuf_used += retsz;

	(void) acopm_log_debug("%s(%s): received %zu bytes just now, %zu total", __func__,
	                       acopm_conn_name(connection), retsz, connection->recvbuf_acqd);
	(void) acopm_log_debug("%s(%s): now have %zu/%zu bytes in recv buffer", __func__,
	                       acopm_conn_name(connection), connection->recvbuf_used, connection->recvbuf_size);

	return true;
}

bool
acopm_conn_writebuf(struct acopm_connection *const restrict connection,
                    const void *const restrict buf, const size_t len)
{
	assert(connection != NULL);
	assert(connection->sendbuf != NULL);

	const size_t sendbuf_free = connection->sendbuf_size - connection->sendbuf_used;

	if (sendbuf_free < len)
	{
		const size_t sendbuf_size_old = connection->sendbuf_size;
		const size_t sendbuf_size_req = connection->sendbuf_used + len;
		const size_t sendbuf_size_new = ((size_t) (((float) sendbuf_size_old) * 1.5F)) + len;

		(void) acopm_log_warning("%s(%s): Not enough free space in send buffer of size %zu bytes; have %zu "
		                         "bytes free, need a buffer of size %zu bytes to queue %zu more bytes, "
		                         "ballooning buffer to %zu bytes", __func__, acopm_conn_name(connection),
		                         sendbuf_size_old, sendbuf_free, sendbuf_size_req, len, sendbuf_size_new);

		if (! acopm_conn_sendbuf_resize(connection, sendbuf_size_new))
			return false;
	}

	(void) memcpy(connection->sendbuf + connection->sendbuf_used, buf, len);
	connection->sendbuf_used += len;

	(void) acopm_log_debug("%s(%s): now have %zu/%zu bytes in send buffer", __func__,
	                       acopm_conn_name(connection), connection->sendbuf_used, connection->sendbuf_size);

	return acopm_conn_ev_mod(connection, EV_WRITE, EV_NONE, connection->ev_rf, connection->ev_sf);
}

void
acopm_conn_free(struct acopm_connection *const restrict connection)
{
	assert(connection != NULL);

	(void) acopm_conn_free_internal(connection);
	(void) free(connection);
}
