/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INC_LIBACOPM_UTILS_BUFCONN_H
#define INC_LIBACOPM_UTILS_BUFCONN_H

#include "acopm-common.h"

#include <limits.h> /* PATH_MAX */
#include <time.h> /* time_t */

#include <netinet/in.h> /* struct sockaddr_* */

#ifndef PATH_MAX
#  define PATH_MAX 4096
#endif

#include <event2/event.h> /* struct event* */

#ifndef EV_NONE
#  define EV_NONE 0
#endif

#ifdef BUILD_WITH_MBEDTLS
#  include <mbedtls/pk.h> /* mbedtls_pk_context */
#  include <mbedtls/ssl.h> /* mbedtls_ssl_* */
#  include <mbedtls/x509_crt.h> /* mbedtls_x509_crt */
#endif

#include "utils/addr.h" /* ACOPM_ADDRSTRLEN */

#define ACOPM_CONN_RECVBUF_LEN_DEFAULT          32768U
#define ACOPM_CONN_SENDBUF_LEN_DEFAULT          32768U

#define ACOPM_CONN_CERTFP_LEN_MIN               20U
#define ACOPM_CONN_CERTFP_LEN_MAX               64U

#define ACOPM_CONN_CERTFP_METHOD_RANGE_MIN      1
#define ACOPM_CONN_CERTFP_METHOD_RANGE_MAX      12

enum acopm_conn_error
{
	ACOPM_CONN_NO_ERROR             = 0,
	ACOPM_CONN_CONNECT_TIMEOUT      = 1,
	ACOPM_CONN_CONNECT_ERROR        = 2,
	ACOPM_CONN_INTERNAL_ERROR       = 3,
	ACOPM_CONN_RECVCB_ERROR         = 4,
	ACOPM_CONN_SENDCB_ERROR         = 5,
	ACOPM_CONN_DATA_TIMEOUT         = 6,
} acopm_attr_packed;

#ifdef BUILD_WITH_MBEDTLS

enum acopm_conn_certfp_method
{
	ACOPM_CONN_CERTFP_METHOD_NONE               =  0,
	ACOPM_CONN_CERTFP_METHOD_CERT_SHA1_HEX      =  1,
	ACOPM_CONN_CERTFP_METHOD_SPKI_SHA1_HEX      =  2,
	ACOPM_CONN_CERTFP_METHOD_CERT_SHA1_B64      =  3,
	ACOPM_CONN_CERTFP_METHOD_SPKI_SHA1_B64      =  4,
	ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_256_HEX  =  5,
	ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_256_HEX  =  6,
	ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_256_B64  =  7,
	ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_256_B64  =  8,
	ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_512_HEX  =  9,
	ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_512_HEX  = 10,
	ACOPM_CONN_CERTFP_METHOD_CERT_SHA2_512_B64  = 11,
	ACOPM_CONN_CERTFP_METHOD_SPKI_SHA2_512_B64  = 12
} acopm_attr_packed;

enum acopm_conn_certfp_length
{
	ACOPM_CONN_CERTFP_LENGTH_UNKNOWN            =   0,
	ACOPM_CONN_CERTFP_LENGTH_SHA1_HEX           =  40,
	ACOPM_CONN_CERTFP_LENGTH_SHA1_B64           =  28,
	ACOPM_CONN_CERTFP_LENGTH_SHA2_256_HEX       =  64,
	ACOPM_CONN_CERTFP_LENGTH_SHA2_256_B64       =  44,
	ACOPM_CONN_CERTFP_LENGTH_SHA2_512_HEX       = 128,
	ACOPM_CONN_CERTFP_LENGTH_SHA2_512_B64       =  88,
} acopm_attr_packed;

struct acopm_tls_conection;
struct acopm_conn_certfp_value;

#endif /* BUILD_WITH_MBEDTLS */

struct acopm_connection;

typedef bool (*acopm_conn_data_func)(struct acopm_connection *);
typedef void (*acopm_conn_error_func)(struct acopm_connection *, enum acopm_conn_error);
typedef int (*acopm_conn_recv_func)(void *, void *, size_t);
typedef int (*acopm_conn_send_func)(void *, const void *, size_t);

struct acopm_connection
{
	void                                    *context;
	struct event_base                       *ev_base;
	struct event                            *ev_r;
	struct event                            *ev_s;
	struct event                            *ev_x;
	unsigned char                           *recvbuf;
	unsigned char                           *sendbuf;
	const struct sockaddr_in                *bind4;
	const struct sockaddr_in6               *bind6;
	acopm_conn_data_func                     ev_rf;
	acopm_conn_data_func                     ev_sf;
	acopm_conn_data_func                     ev_rfs;
	acopm_conn_data_func                     ev_sfs;
	acopm_conn_error_func                    ev_ef;
	acopm_conn_recv_func                     net_recv_func;
	acopm_conn_send_func                     net_send_func;

#ifdef BUILD_WITH_MBEDTLS
	struct acopm_tls_conection              *tls;
#endif

	char                                     name[256];

	size_t                                   recvbuf_size;
	size_t                                   recvbuf_used;
	size_t                                   recvbuf_acqd;
	size_t                                   sendbuf_size;
	size_t                                   sendbuf_orig;
	size_t                                   sendbuf_used;

	time_t                                   connect_timeout;
	time_t                                   data_timeout;

	int                                      socket_fd;
	int                                      net_errno;

	short                                    ev_e;
	short                                    ev_es;

	bool                                     connected:1;
	bool                                     ev_break:1;
	bool                                     nonblock:1;
	bool                                     stashed:1;

#ifdef BUILD_WITH_MBEDTLS
	bool                                     use_tls:1;
#endif

} acopm_attr_packed;

#ifdef BUILD_WITH_MBEDTLS

struct acopm_tls_conection
{
	mbedtls_x509_crt                         tls_crt;
	mbedtls_pk_context                       tls_key;
	mbedtls_ssl_config                       tls_cfg;
	mbedtls_ssl_context                      tls_ctx;

	const struct acopm_conn_certfp_value    *certfp_values;

	enum acopm_conn_certfp_method            certfp_method;

	char                                     cc_file[PATH_MAX + 1];
	char                                     ck_file[PATH_MAX + 1];

	bool                                     cert_verification_success:1;
} acopm_attr_packed;

struct acopm_conn_certfp_value
{
	struct acopm_conn_certfp_value          *next;
	char                                     value[(ACOPM_CONN_CERTFP_LEN_MAX * 2) + 1];
} acopm_attr_packed;

#endif

struct acopm_connection *acopm_conn_new(const char *, void *) acopm_attr_malloc;
void acopm_conn_free(struct acopm_connection *);

void acopm_conn_set_ev_break(struct acopm_connection *, bool);
void acopm_conn_set_nonblock(struct acopm_connection *, bool, time_t);
void acopm_conn_set_data_cb(struct acopm_connection *, short, acopm_conn_data_func, acopm_conn_data_func);
void acopm_conn_set_error_cb(struct acopm_connection *, acopm_conn_error_func);
void acopm_conn_set_data_timeout(struct acopm_connection *, time_t);
void acopm_conn_set_bufsz(struct acopm_connection *, size_t, size_t);

void acopm_conn_set_bind4(struct acopm_connection *, const struct sockaddr_in *);
void acopm_conn_set_bind6(struct acopm_connection *, const struct sockaddr_in6 *);

#ifdef BUILD_WITH_MBEDTLS
bool acopm_conn_set_tls(struct acopm_connection *, bool, enum acopm_conn_certfp_method,
                        const struct acopm_conn_certfp_value *);
void acopm_conn_set_tls_cert(struct acopm_connection *, const char *, const char *);
#endif

bool acopm_conn_do_connect(struct acopm_connection *, struct event_base *, struct sockaddr *, socklen_t);

bool acopm_conn_get_sockaddr(struct acopm_connection *,
                             char al[static ACOPM_ADDRSTRLEN], char ar[static ACOPM_ADDRSTRLEN]);

bool acopm_conn_ev_mod(struct acopm_connection *, short, short, acopm_conn_data_func, acopm_conn_data_func);

bool acopm_attr_warn_unused acopm_conn_readbuf(struct acopm_connection *, acopm_conn_data_func);
bool acopm_attr_warn_unused acopm_conn_writebuf(struct acopm_connection *, const void *, size_t);

#endif /* !INC_LIBACOPM_UTILS_BUFCONN_H */
