/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <errno.h> /* errno */
#include <string.h> /* memcmp(), strerror() */

#include <arpa/inet.h> /* inet_pton() */
#include <netinet/in.h> /* INET6_ADDRSTRLEN */
#include <sys/socket.h> /* AF_INET, AF_INET6 */

#include "utils/cidr.h" /* decls for own functions */
#include "utils/log.h" /* acopm_log_error() */

struct acopm_cidr_address *
acopm_cidr_new(const char *const restrict addr)
{
	assert(addr != NULL);

	char addr_copy[INET6_ADDRSTRLEN + 4];

	if (! acopm_strset_buf(addr_copy, addr))
	{
		(void) acopm_log_error("%s: argument too large", __func__);
		return NULL;
	}

	struct acopm_cidr_address *const newcidr = calloc(1, sizeof *newcidr);

	if (! newcidr)
	{
		(void) acopm_log_error("%s: malloc: %s", __func__, strerror(errno));
		return NULL;
	}

	char *prefix = strchr(addr_copy, '/');

	if (prefix)
	{
		*prefix++ = 0x00;

		if (*prefix == 0x00)
		{
			(void) acopm_log_error("%s: missing prefix length", __func__);
			(void) acopm_cidr_free(newcidr);
			return NULL;
		}

		for (size_t i = 0; prefix[i] != 0x00; i++)
		{
			if (prefix[i] < 0x30 || prefix[i] > 0x39)
			{
				(void) acopm_log_error("%s: invalid prefix length '%s'", __func__, prefix);
				(void) acopm_cidr_free(newcidr);
				return NULL;
			}
		}
	}

	const char *const v4 = strchr(addr_copy, '.');
	const char *const v6 = strchr(addr_copy, ':');

	if (! v4 && ! v6)
	{
		(void) acopm_log_error("%s: invalid address '%s'", __func__, addr_copy);
		(void) acopm_cidr_free(newcidr);
		return NULL;
	}
	else if (v6 && inet_pton(AF_INET6, addr_copy, newcidr->address) == 1)
	{
		newcidr->family = AF_INET6;
	}
	else if (v4 && inet_pton(AF_INET, addr_copy, newcidr->address) == 1)
	{
		newcidr->family = AF_INET;
	}
	else
	{
		(void) acopm_log_error("%s: inet_pton: %s", __func__, strerror(errno));
		(void) acopm_log_error("%s: invalid address '%s'", __func__, addr_copy);
		(void) acopm_cidr_free(newcidr);
		return NULL;
	}

	newcidr->netsize = ((newcidr->family == AF_INET) ? INET_ADDRLEN : INET6_ADDRLEN) * 8;

	if (prefix)
	{
		char *endptr = NULL;
		const unsigned long int val = strtoul(prefix, &endptr, 10);

		assert(endptr != NULL);

		if (! endptr || *endptr)
		{
			(void) acopm_log_error("%s: invalid prefix length (at '%s')",
			                       __func__, ((endptr) ? endptr : "<unknown>"));

			(void) acopm_cidr_free(newcidr);
			return NULL;
		}
		if (val < 1 || val > (unsigned long int) newcidr->netsize)
		{
			(void) acopm_log_error("%s: invalid prefix length %lu", __func__, val);
			(void) acopm_cidr_free(newcidr);
			return NULL;
		}

		newcidr->netsize = (uint8_t) val;
	}

	return newcidr;
}

void
acopm_cidr_free(struct acopm_cidr_address *const restrict cidr)
{
	(void) free(cidr);
}

bool
acopm_cidr_match(const char *const restrict addr, const struct acopm_cidr_address *const restrict cidr)
{
	assert(addr != NULL);
	assert(cidr != NULL);
	assert(cidr->family == AF_INET || cidr->family == AF_INET6);

	char addr_copy[INET6_ADDRSTRLEN + 4];

	if (! acopm_strset_buf(addr_copy, addr))
	{
		(void) acopm_log_error("%s: argument too large", __func__);
		return false;
	}

	char *const prefix = strchr(addr_copy, '/');

	if (prefix)
		*prefix = 0x00;

	uint8_t addr_raw[INET6_ADDRLEN];

	if (inet_pton(cidr->family, addr_copy, addr_raw) != 1)
		return false;

	const uint8_t *const addr_cmp = cidr->address;
	const size_t octs = (size_t) (cidr->netsize / 8);

	if (octs == 0 || memcmp(addr_cmp, addr_raw, octs) == 0)
	{
		const uint8_t rem = (uint8_t) ~(((uint16_t) (1 << (8 - (cidr->netsize % 8)))) - 1);

		if ((addr_cmp[octs] & rem) == (addr_raw[octs] & rem))
			return true;
	}

	return false;
}
