/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <errno.h> /* errno */
#include <fcntl.h> /* fcntl() */
#include <string.h> /* strerror() */
#include <time.h> /* time(), localtime(), strftime(), time_t, struct tm */

#include "utils/log.h" /* decls for own functions */

#define TIMESTAMP_LEN 24

static FILE *acopm_log_file = NULL;
static unsigned int acopm_log_mask = 0;

static bool acopm_log_mask_set = false;
static bool acopm_log_file_set = false;

static inline const char * acopm_attr_inline
acopm_log_get_level_str(const unsigned int level)
{
	switch (level)
	{
		case ACOPM_LOGLVL_DEBUG:
			return "DEBUG";

		case ACOPM_LOGLVL_TRAFFIC:
			return "TRAFFIC";

		case ACOPM_LOGLVL_INFO:
			return "INFO";

		case ACOPM_LOGLVL_NOTICE:
			return "NOTICE";

		case ACOPM_LOGLVL_WARNING:
			return "WARNING";

		case ACOPM_LOGLVL_ERROR:
			return "ERROR";

		case ACOPM_LOGLVL_CRITICAL:
			return "CRITICAL";
	}

	abort();
}

static inline void acopm_attr_inline acopm_attr_printf(2, 0)
acopm_log_internal(const unsigned int level, const char *const restrict format, va_list argv)
{
	const time_t current = time(NULL);
	const struct tm *timep = NULL;
	char timestamp[TIMESTAMP_LEN];

	if (current == ((time_t) -1) || ! (timep = localtime(&current)))
		goto print_message;

	if (! strftime(timestamp, sizeof timestamp, "%Y-%m-%d %H:%M:%S", timep))
		goto print_message;

	(void) fprintf(acopm_log_file, "%s ", timestamp);

print_message:
	(void) fprintf(acopm_log_file, "[%10s] ", acopm_log_get_level_str(level));
	(void) vfprintf(acopm_log_file, format, argv);
	(void) fprintf(acopm_log_file, "\n");
	(void) fflush(acopm_log_file);
}

bool
acopm_log_init(void)
{
	acopm_log_file = stderr;
	acopm_log_mask = ACOPM_LOGLVL_DEFAULT;

	return true;
}

bool
acopm_log_file_is_set(void)
{
	return acopm_log_file_set;
}

bool
acopm_log_mask_is_set(void)
{
	return acopm_log_mask_set;
}

bool
acopm_log_set_file(const char *const restrict filename)
{
	if (acopm_log_file_set)
		return false;

	(void) acopm_log_info("Switching to log file '%s'", filename);

	FILE *const ptr = fopen(filename, "a");

	if (! ptr)
	{
		(void) acopm_log_error("%s: fopen: %s", __func__, strerror(errno));
		return false;
	}

	const int fd = fileno(ptr);

	if (fd == -1)
	{
		(void) acopm_log_error("%s: fileno: %s", __func__, strerror(errno));
		(void) fclose(ptr);
		return false;
	}

	const int flags = fcntl(fd, F_GETFD, 0);

	if (flags == -1 || fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1)
	{
		(void) acopm_log_error("%s: fcntl: %s", __func__, strerror(errno));
		(void) fclose(ptr);
		return false;
	}

	(void) fflush(acopm_log_file);
	acopm_log_file = ptr;
	return true;
}

void
acopm_log_set_mask(const unsigned int mask)
{
	if (acopm_log_mask_set)
		return;

	acopm_log_mask = mask;
	acopm_log_mask_set = true;
}

void
acopm_log_flush(void)
{
	(void) fflush(acopm_log_file);
}

void
acopm_log_free(void)
{
	if (! acopm_log_file || acopm_log_file == stderr)
		return;

	(void) fflush(acopm_log_file);
	(void) fclose(acopm_log_file);

	acopm_log_file = stderr;
}

void
acopm_log_critical(const char *const restrict format, ...)
{
	if (! (acopm_log_mask & ACOPM_LOGLVL_CRITICAL))
		return;

	va_list argv;
	va_start(argv, format);
	(void) acopm_log_internal(ACOPM_LOGLVL_CRITICAL, format, argv);
	va_end(argv);
}

void
acopm_log_error(const char *const restrict format, ...)
{
	if (! (acopm_log_mask & ACOPM_LOGLVL_ERROR))
		return;

	va_list argv;
	va_start(argv, format);
	(void) acopm_log_internal(ACOPM_LOGLVL_ERROR, format, argv);
	va_end(argv);
}

void
acopm_log_warning(const char *const restrict format, ...)
{
	if (! (acopm_log_mask & ACOPM_LOGLVL_WARNING))
		return;

	va_list argv;
	va_start(argv, format);
	(void) acopm_log_internal(ACOPM_LOGLVL_WARNING, format, argv);
	va_end(argv);
}

void
acopm_log_notice(const char *const restrict format, ...)
{
	if (! (acopm_log_mask & ACOPM_LOGLVL_NOTICE))
		return;

	va_list argv;
	va_start(argv, format);
	(void) acopm_log_internal(ACOPM_LOGLVL_NOTICE, format, argv);
	va_end(argv);
}

void
acopm_log_info(const char *const restrict format, ...)
{
	if (! (acopm_log_mask & ACOPM_LOGLVL_INFO))
		return;

	va_list argv;
	va_start(argv, format);
	(void) acopm_log_internal(ACOPM_LOGLVL_INFO, format, argv);
	va_end(argv);
}

void
acopm_log_traffic(const char *const restrict format, ...)
{
	if (! (acopm_log_mask & ACOPM_LOGLVL_TRAFFIC))
		return;

	va_list argv;
	va_start(argv, format);
	(void) acopm_log_internal(ACOPM_LOGLVL_TRAFFIC, format, argv);
	va_end(argv);
}

#ifndef NDEBUG
void
acopm_log_debug(const char *const restrict format, ...)
{
	if (! (acopm_log_mask & ACOPM_LOGLVL_DEBUG))
		return;

	va_list argv;
	va_start(argv, format);
	(void) acopm_log_internal(ACOPM_LOGLVL_DEBUG, format, argv);
	va_end(argv);
}
#endif
