/*
 * SipHash-2-4 reference C implementation
 *
 * Copyright (C) 2012-2016 Jean-Philippe Aumasson
 *               <jeanphilippe.aumasson@gmail.com>
 *
 * Copyright (C) 2012-2014 Daniel J. Bernstein <djb@cr.yp.to>
 *
 * To the extent possible under law, the author(s) have dedicated all
 * copyright and related and neighboring rights to this software to the
 * public domain worldwide.
 *
 * THIS SOFTWARE IS DISTRIBUTED WITHOUT ANY WARRANTY.
 *
 * Modified by Aaron M. D. Jones <aaronmdjones@gmail.com> (2017):
 *
 * * Separate the digest algorithm from its key parameter (allow the key
 *   to be generated once, set, and then never used in function calls
 *   again).
 */

#include "acopm-common.h"

#include <assert.h> /* assert() */
#include <endian.h> /* htole64() */
#include <inttypes.h> /* PRIX64 */

#include "utils/log.h" /* acopm_log_*() */
#include "utils/siphash24.h" /* decls for own functions */

#define ROTL(x, y)      ((uint64_t) (((x) << (y)) | ((x) >> (0x40 - (y)))))

#define SIPROUND(x)                                                             \
    do {                                                                        \
        x[0x00] += x[0x01]; x[0x01] = ROTL(x[0x01], 0x0D);                      \
        x[0x01] ^= x[0x00]; x[0x00] = ROTL(x[0x00], 0x20);                      \
        x[0x02] += x[0x03]; x[0x03] = ROTL(x[0x03], 0x10); x[0x03] ^= x[0x02];  \
        x[0x00] += x[0x03]; x[0x03] = ROTL(x[0x03], 0x15); x[0x03] ^= x[0x00];  \
        x[0x02] += x[0x01]; x[0x01] = ROTL(x[0x01], 0x11);                      \
        x[0x01] ^= x[0x02]; x[0x02] = ROTL(x[0x02], 0x20);                      \
    } while(0)

static bool acopm_siphash24_key_set = false;

static uint64_t acopm_siphash24_iv[4] = {

	UINT64_C(0x736F6D6570736575), UINT64_C(0x646F72616E646F6D),
	UINT64_C(0x6C7967656E657261), UINT64_C(0x7465646279746573),
};

bool
acopm_siphash24_setkey(const uint8_t key[static SIPHASH24_KEY_LENGTH])
{
	if (acopm_siphash24_key_set)
		return false;

	for (size_t i = 0x00; i < 0x04; i++)
	{
		acopm_siphash24_iv[i] ^= htole64(*((const uint64_t *) &key[(i % 0x02) * 0x08]));

		(void) acopm_log_debug("%s: iv[%zu]: 0x%16" PRIX64, __func__, i, acopm_siphash24_iv[i]);
	}

	acopm_siphash24_key_set = true;
	return true;
}

uint64_t
acopm_siphash24_digest(const void *const restrict ptr, const size_t len)
{
	assert(acopm_siphash24_key_set == true);
	assert(ptr != NULL);

	const uint8_t *buf = (const uint8_t *) ptr;
	const uint8_t *const end = buf + len - (len % 0x08);

	uint64_t v[0x04];
	uint64_t x = (((uint64_t) len) << 0x38);

	(void) memcpy(v, acopm_siphash24_iv, sizeof v);

	for ( ; buf < end; buf += 0x08)
	{
		const uint64_t m = htole64(*((const uint64_t *) buf));

		v[0x03] ^= m;

		SIPROUND(v);
		SIPROUND(v);

		v[0x00] ^= m;
	}

	for (uint64_t i = (len & 0x07); i != 0x00; i--)
		x |= (((uint64_t) buf[i - 0x01]) << (0x08 * (i - 0x01)));

	v[0x03] ^= x;

	SIPROUND(v);
	SIPROUND(v);

	v[0x00] ^= x;
	v[0x02] ^= 0xFF;

	SIPROUND(v);
	SIPROUND(v);
	SIPROUND(v);
	SIPROUND(v);

	return htole64(v[0x00] ^ v[0x01] ^ v[0x02] ^ v[0x03]);
}
