/*
 * Copyright (C) 2017 Aaron M. D. Jones <aaronmdjones@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.
 *
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "acopm-common.h"

#include <inttypes.h> /* PRIX64 */

#include "tests.h" /* EXIT_AUTOMAKE_CHECK_HARD_FAILURE */
#include "utils/log.h" /* acopm_log_*() */
#include "utils/siphash24.h" /* acopm_siphash24_*() */

static const uint8_t vectors[0x40][0x08] = {

	{ 0x31, 0x0E, 0x0E, 0xDD, 0x47, 0xDB, 0x6F, 0x72 }, { 0xFD, 0x67, 0xDC, 0x93, 0xC5, 0x39, 0xF8, 0x74 },
	{ 0x5A, 0x4F, 0xA9, 0xD9, 0x09, 0x80, 0x6C, 0x0D }, { 0x2D, 0x7E, 0xFB, 0xD7, 0x96, 0x66, 0x67, 0x85 },
	{ 0xB7, 0x87, 0x71, 0x27, 0xE0, 0x94, 0x27, 0xCF }, { 0x8D, 0xA6, 0x99, 0xCD, 0x64, 0x55, 0x76, 0x18 },
	{ 0xCE, 0xE3, 0xFE, 0x58, 0x6E, 0x46, 0xC9, 0xCB }, { 0x37, 0xD1, 0x01, 0x8B, 0xF5, 0x00, 0x02, 0xAB },
	{ 0x62, 0x24, 0x93, 0x9A, 0x79, 0xF5, 0xF5, 0x93 }, { 0xB0, 0xE4, 0xA9, 0x0B, 0xDF, 0x82, 0x00, 0x9E },
	{ 0xF3, 0xB9, 0xDD, 0x94, 0xC5, 0xBB, 0x5D, 0x7A }, { 0xA7, 0xAD, 0x6B, 0x22, 0x46, 0x2F, 0xB3, 0xF4 },
	{ 0xFB, 0xE5, 0x0E, 0x86, 0xBC, 0x8F, 0x1E, 0x75 }, { 0x90, 0x3D, 0x84, 0xC0, 0x27, 0x56, 0xEA, 0x14 },
	{ 0xEE, 0xF2, 0x7A, 0x8E, 0x90, 0xCA, 0x23, 0xF7 }, { 0xE5, 0x45, 0xBE, 0x49, 0x61, 0xCA, 0x29, 0xA1 },
	{ 0xDB, 0x9B, 0xC2, 0x57, 0x7F, 0xCC, 0x2A, 0x3F }, { 0x94, 0x47, 0xBE, 0x2C, 0xF5, 0xE9, 0x9A, 0x69 },
	{ 0x9C, 0xD3, 0x8D, 0x96, 0xF0, 0xB3, 0xC1, 0x4B }, { 0xBD, 0x61, 0x79, 0xA7, 0x1D, 0xC9, 0x6D, 0xBB },
	{ 0x98, 0xEE, 0xA2, 0x1A, 0xF2, 0x5C, 0xD6, 0xBE }, { 0xC7, 0x67, 0x3B, 0x2E, 0xB0, 0xCB, 0xF2, 0xD0 },
	{ 0x88, 0x3E, 0xA3, 0xE3, 0x95, 0x67, 0x53, 0x93 }, { 0xC8, 0xCE, 0x5C, 0xCD, 0x8C, 0x03, 0x0C, 0xA8 },
	{ 0x94, 0xAF, 0x49, 0xF6, 0xC6, 0x50, 0xAD, 0xB8 }, { 0xEA, 0xB8, 0x85, 0x8A, 0xDE, 0x92, 0xE1, 0xBC },
	{ 0xF3, 0x15, 0xBB, 0x5B, 0xB8, 0x35, 0xD8, 0x17 }, { 0xAD, 0xCF, 0x6B, 0x07, 0x63, 0x61, 0x2E, 0x2F },
	{ 0xA5, 0xC9, 0x1D, 0xA7, 0xAC, 0xAA, 0x4D, 0xDE }, { 0x71, 0x65, 0x95, 0x87, 0x66, 0x50, 0xA2, 0xA6 },
	{ 0x28, 0xEF, 0x49, 0x5C, 0x53, 0xA3, 0x87, 0xAD }, { 0x42, 0xC3, 0x41, 0xD8, 0xFA, 0x92, 0xD8, 0x32 },
	{ 0xCE, 0x7C, 0xF2, 0x72, 0x2F, 0x51, 0x27, 0x71 }, { 0xE3, 0x78, 0x59, 0xF9, 0x46, 0x23, 0xF3, 0xA7 },
	{ 0x38, 0x12, 0x05, 0xBB, 0x1A, 0xB0, 0xE0, 0x12 }, { 0xAE, 0x97, 0xA1, 0x0F, 0xD4, 0x34, 0xE0, 0x15 },
	{ 0xB4, 0xA3, 0x15, 0x08, 0xBE, 0xFF, 0x4D, 0x31 }, { 0x81, 0x39, 0x62, 0x29, 0xF0, 0x90, 0x79, 0x02 },
	{ 0x4D, 0x0C, 0xF4, 0x9E, 0xE5, 0xD4, 0xDC, 0xCA }, { 0x5C, 0x73, 0x33, 0x6A, 0x76, 0xD8, 0xBF, 0x9A },
	{ 0xD0, 0xA7, 0x04, 0x53, 0x6B, 0xA9, 0x3E, 0x0E }, { 0x92, 0x59, 0x58, 0xFC, 0xD6, 0x42, 0x0C, 0xAD },
	{ 0xA9, 0x15, 0xC2, 0x9B, 0xC8, 0x06, 0x73, 0x18 }, { 0x95, 0x2B, 0x79, 0xF3, 0xBC, 0x0A, 0xA6, 0xD4 },
	{ 0xF2, 0x1D, 0xF2, 0xE4, 0x1D, 0x45, 0x35, 0xF9 }, { 0x87, 0x57, 0x75, 0x19, 0x04, 0x8F, 0x53, 0xA9 },
	{ 0x10, 0xA5, 0x6C, 0xF5, 0xDF, 0xCD, 0x9A, 0xDB }, { 0xEB, 0x75, 0x09, 0x5C, 0xCD, 0x98, 0x6C, 0xD0 },
	{ 0x51, 0xA9, 0xCB, 0x9E, 0xCB, 0xA3, 0x12, 0xE6 }, { 0x96, 0xAF, 0xAD, 0xFC, 0x2C, 0xE6, 0x66, 0xC7 },
	{ 0x72, 0xFE, 0x52, 0x97, 0x5A, 0x43, 0x64, 0xEE }, { 0x5A, 0x16, 0x45, 0xB2, 0x76, 0xD5, 0x92, 0xA1 },
	{ 0xB2, 0x74, 0xCB, 0x8E, 0xBF, 0x87, 0x87, 0x0A }, { 0x6F, 0x9B, 0xB4, 0x20, 0x3D, 0xE7, 0xB3, 0x81 },
	{ 0xEA, 0xEC, 0xB2, 0xA3, 0x0B, 0x22, 0xA8, 0x7F }, { 0x99, 0x24, 0xA4, 0x3C, 0xC1, 0x31, 0x57, 0x24 },
	{ 0xBD, 0x83, 0x8D, 0x3A, 0xAF, 0xBF, 0x8D, 0xB7 }, { 0x0B, 0x1A, 0x2A, 0x32, 0x65, 0xD5, 0x1A, 0xEA },
	{ 0x13, 0x50, 0x79, 0xA3, 0x23, 0x1C, 0xE6, 0x60 }, { 0x93, 0x2B, 0x28, 0x46, 0xE4, 0xD7, 0x06, 0x66 },
	{ 0xE1, 0x91, 0x5F, 0x5C, 0xB1, 0xEC, 0xA4, 0x6C }, { 0xF3, 0x25, 0x96, 0x5C, 0xA1, 0x6D, 0x62, 0x9F },
	{ 0x57, 0x5F, 0xF2, 0x8E, 0x60, 0x38, 0x1B, 0xE5 }, { 0x72, 0x45, 0x06, 0xEB, 0x4C, 0x32, 0x8A, 0x95 },
};

static inline bool
acopm_siphash24_init(void)
{
	uint8_t sipkey24[SIPHASH24_KEY_LENGTH];

	for (uint8_t i = 0x00; i < sizeof sipkey24; i++)
		sipkey24[i] = i;

	return acopm_siphash24_setkey(sipkey24);
}

static inline bool
acopm_siphash24_test(void)
{
	uint8_t data[0x40];

	bool ret = true;

	for (size_t i = 0x00; i < 0x40; i++)
	{
		data[i] = (uint8_t) i;

		const uint64_t res = acopm_siphash24_digest(data, i);

		if (memcmp(&res, vectors[i], sizeof res) != 0)
		{
			(void) acopm_log_critical("%s: test %zu does not match, got 0x%16" PRIX64, __func__, i, res);
			ret = false;
		}
	}

	return ret;
}

int
main(void)
{
	if (! acopm_log_init())
		return EXIT_AUTOMAKE_CHECK_HARD_FAILURE;

	if (! acopm_siphash24_init())
		return EXIT_AUTOMAKE_CHECK_HARD_FAILURE;

	if (! acopm_siphash24_test())
		return EXIT_FAILURE;

	return EXIT_SUCCESS;
}
